# Copyright 2011 Dan Smith <dsmith@danplanet.com>
# Copyright 2012 Tom Hayward <tom@tomh.us>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp.drivers import yaesu_clone
from chirp import chirp_common, directory, errors, bitwise

MEM_FORMAT = """
struct mem_struct {
  u8 unknown1:2,
     half_deviation:1,
     cpu_shift:1,
     unknown2:1,
     sp_rx1:1,
     unknown3:1,
     sp_rx2:1;
  u8 unknown4:4,
     tuning_step:4;
  bbcd freq[3];
  u8 icon:6,
     mode:2;
  char name[8];
  bbcd offset[3];
  u8 tmode:4,
     power:2,
     duplex:2;
  u8 unknown5:2,
     tone:6;
  u8 unknown6:1,
     dtcs:7;
  u8 unknown7;
  u8 unknown8;
};

struct flag_struct {
  u8 zeros:4,
     pskip:1,
     skip:1,
     visible:1,
     used:1;
};

#seekto 0x002A;
struct {
  u8 current_member;
} bank_used[5];

#seekto 0x0032;
struct {
  struct {
    u8 status;
    u8 channel;
  } members[24];
} bank_groups[5];

#seekto 0x012A;
struct flag_struct flag[220];
struct flag_struct specialflag[20];

#seekto 0x026A;
struct mem_struct memory[220];
struct mem_struct special[50];

#seekto 0x1D03;
u8 current_bank;
"""

TMODES = ["", "Tone", "TSQL", "DTCS"]
DUPLEX = ["", "-", "+", "split"]
MODES = ["FM", "AM", "WFM"]
STEPS = [5.0, 10.0, 12.5, 15.0, 20.0, 25.0, 50.0, 100.0]
POWER_LEVELS = [chirp_common.PowerLevel("Hi", watts=5.00),
                chirp_common.PowerLevel("L3", watts=2.50),
                chirp_common.PowerLevel("L2", watts=1.00),
                chirp_common.PowerLevel("L1", watts=0.05)]
SPECIALS = ["%s%d" % (c, i + 1) for i in range(0, 10) for c in ('L', 'U')]


class VX5BankModel(chirp_common.BankModel):
    def get_num_mappings(self):
        return 5

    def get_mappings(self):
        banks = []
        for i in range(0, self.get_num_mappings()):
            bank = chirp_common.Bank(self, "%i" % (i+1), "MG%i" % (i+1))
            bank.index = i
            banks.append(bank)
        return banks

    def add_memory_to_mapping(self, memory, bank):
        _members = self._radio._memobj.bank_groups[bank.index].members
        _bank_used = self._radio._memobj.bank_used[bank.index]
        for i in range(0, len(_members)):
            if _members[i].status == 0xFF:
                # LOG.debug("empty found, inserting %d at %d" %
                #           (memory.number, i))
                if self._radio._memobj.current_bank == 0xFF:
                    self._radio._memobj.current_bank = bank.index
                _members[i].status = 0x00
                _members[i].channel = memory.number - 1
                _bank_used.current_member = i
                return True
        raise Exception(_("{bank} is full").format(bank=bank))

    def remove_memory_from_mapping(self, memory, bank):
        _members = self._radio._memobj.bank_groups[bank.index].members
        _bank_used = self._radio._memobj.bank_used[bank.index]

        found = False
        remaining_members = 0
        for i in range(0, len(_members)):
            if _members[i].status == 0x00:
                if _members[i].channel == (memory.number - 1):
                    _members[i].status = 0xFF
                    found = True
                else:
                    remaining_members += 1

        if not found:
            raise Exception(_("Memory {num} not in "
                              "bank {bank}").format(num=memory.number,
                                                    bank=bank))
        if not remaining_members:
            _bank_used.current_member = 0xFF

    def get_mapping_memories(self, bank):
        memories = []

        _members = self._radio._memobj.bank_groups[bank.index].members
        _bank_used = self._radio._memobj.bank_used[bank.index]

        if _bank_used.current_member == 0xFF:
            return memories

        for member in _members:
            if member.status == 0xFF:
                continue
            memories.append(self._radio.get_memory(member.channel+1))
        return memories

    def get_memory_mappings(self, memory):
        banks = []
        for bank in self.get_mappings():
            if memory.number in [x.number for x in
                                 self.get_mapping_memories(bank)]:
                banks.append(bank)
        return banks


@directory.register
class VX5Radio(yaesu_clone.YaesuCloneModeRadio):
    """Yaesu VX-5"""
    BAUD_RATE = 9600
    VENDOR = "Yaesu"
    MODEL = "VX-5"

    _model = ""
    _memsize = 8123
    _block_lengths = [10, 16, 8097]
    _block_size = 8

    def _checksums(self):
        return [yaesu_clone.YaesuChecksum(0x0000, 0x1FB9, 0x1FBA)]

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.can_odd_split = True
        rf.has_bank = True
        rf.has_ctone = False
        rf.has_dtcs_polarity = False
        rf.valid_modes = MODES + ["NFM"]
        rf.valid_tmodes = TMODES
        rf.valid_tuning_steps = STEPS
        rf.valid_duplexes = DUPLEX
        rf.memory_bounds = (1, 220)
        rf.valid_bands = [(500000,    16000000),
                          (48000000,  729000000),
                          (800000000, 999000000)]
        rf.valid_skips = ["", "S", "P"]
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_name_length = 8
        rf.valid_characters = chirp_common.CHARSET_ASCII
        rf.valid_special_chans = SPECIALS

        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number-1])

    def get_memory(self, number):
        mem = chirp_common.Memory()
        if isinstance(number, str):
            mem.number = len(self._memobj.memory) + SPECIALS.index(number) + 1
            mem.extd_number = number
            _mem = self._memobj.special[mem.number -
                                        len(self._memobj.memory) - 1]
            _flg = self._memobj.specialflag[mem.number -
                                            len(self._memobj.memory) - 1]
        elif number > len(self._memobj.memory):
            mem.number = number
            mem.extd_number = SPECIALS[number - len(self._memobj.memory) - 1]
            _mem = self._memobj.special[mem.number -
                                        len(self._memobj.memory) - 1]
            _flg = self._memobj.specialflag[mem.number -
                                            len(self._memobj.memory) - 1]
        else:
            mem.number = number
            _mem = self._memobj.memory[mem.number - 1]
            _flg = self._memobj.flag[mem.number - 1]

        if not _flg.visible:
            mem.empty = True
        if not _flg.used:
            mem.empty = True
            return mem

        mem.freq = chirp_common.fix_rounded_step(int(_mem.freq) * 1000)
        mem.duplex = DUPLEX[_mem.duplex]
        mem.name = self.filter_name(str(_mem.name).rstrip())
        mem.mode = MODES[_mem.mode]
        if mem.mode == "FM" and _mem.half_deviation:
            mem.mode = "NFM"
        mem.tuning_step = STEPS[_mem.tuning_step]
        mem.offset = int(_mem.offset) * 1000
        mem.power = POWER_LEVELS[3 - _mem.power]
        mem.tmode = TMODES[_mem.tmode & 0x3]  # masked so bad mems can be read
        if mem.duplex == "split":
            mem.offset = chirp_common.fix_rounded_step(mem.offset)
        mem.rtone = mem.ctone = chirp_common.OLD_TONES[_mem.tone]
        mem.dtcs = chirp_common.DTCS_CODES[_mem.dtcs]

        mem.skip = _flg.pskip and "P" or _flg.skip and "S" or ""

        return mem

    def set_memory(self, mem):
        if mem.number > len(self._memobj.memory):
            _mem = self._memobj.special[mem.number -
                                        len(self._memobj.memory) - 1]
            _flg = self._memobj.specialflag[mem.number -
                                            len(self._memobj.memory) - 1]
        else:
            _mem = self._memobj.memory[mem.number - 1]
            _flg = self._memobj.flag[mem.number - 1]

        # initialize new channel to safe defaults
        if not mem.empty and not _flg.used:
            _flg.used = True
            _mem.unknown1 = 0x00
            _mem.unknown2 = 0x0
            _mem.unknown3 = 0x0
            _mem.unknown4 = 0x0000
            _mem.unknown5 = 0x0
            _mem.unknown6 = 0x0
            _mem.unknown7 = 0x00
            _mem.unknown8 = 0x00
            _mem.cpu_shift = 0x0
            _mem.sp_rx1 = 0x0
            _mem.sp_rx2 = 0x0

            _mem.icon = 12  # file cabinet icon

        if mem.empty and _flg.used and not _flg.visible:
            _flg.used = False
            return
        _flg.visible = not mem.empty
        if mem.empty:
            self._wipe_memory_banks(mem)
            return

        _mem.freq = int(mem.freq / 1000)
        _mem.duplex = DUPLEX.index(mem.duplex)
        _mem.name = mem.name.ljust(8)
        if mem.mode == "NFM":
            _mem.mode = MODES.index("FM")
            _mem.half_deviation = 1
        else:
            _mem.mode = MODES.index(mem.mode)
            _mem.half_deviation = 0
        _mem.tuning_step = STEPS.index(mem.tuning_step)
        _mem.offset = int(mem.offset / 1000)
        if mem.power:
            _mem.power = 3 - POWER_LEVELS.index(mem.power)
        else:
            _mem.power = 0
        _mem.tmode = TMODES.index(mem.tmode)
        try:
            _mem.tone = chirp_common.OLD_TONES.index(mem.rtone)
        except ValueError:
            raise errors.UnsupportedToneError(
                ("This radio does not support tone %s" % mem.rtone))
        _mem.dtcs = chirp_common.DTCS_CODES.index(mem.dtcs)

        _flg.skip = mem.skip == "S"
        _flg.pskip = mem.skip == "P"

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.pre_download = _(
            "1. Turn radio off.\n"
            "2. Connect cable to MIC/EAR jack.\n"
            "3. Press and hold in the [F/W] key while turning the radio on\n"
            "    (\"CLONE\" will appear on the display).\n"
            "4. <b>After clicking OK</b>, press the [VFO(DW)SC] key to"
            " receive\n"
            "    the image from the radio.\n")
        rp.pre_upload = _(
            "1. Turn radio off.\n"
            "2. Connect cable to MIC/EAR jack.\n"
            "3. Press and hold in the [F/W] key while turning the radio on\n"
            "    (\"CLONE\" will appear on the display).\n"
            "4. Press the [MR(SKP)SC] key (\"CLONE WAIT\" will appear\n"
            "    on the LCD).\n"
            "5. Click OK to send image to radio.\n")
        return rp

    @classmethod
    def match_model(cls, filedata, filename):
        return len(filedata) == cls._memsize

    def get_bank_model(self):
        return VX5BankModel(self)



# Copyright 2010 Dan Smith <dsmith@danplanet.com>
# Copyright 2017 Wade Simmons <wade@wades.im>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from chirp.drivers import yaesu_clone, ft1d
from chirp import chirp_common, directory, bitwise
from chirp.settings import RadioSettings

LOG = logging.getLogger(__name__)

POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=5),
                chirp_common.PowerLevel("Mid", watts=30),
                chirp_common.PowerLevel("Hi", watts=65)]

TMODES = ["", "Tone", "TSQL", "DTCS", "TSQL-R", None, None, "Pager", "Cross"]
CROSS_MODES = [None, "DTCS->", "Tone->DTCS", "DTCS->Tone"]

MODES = ["FM", "NFM"]
STEPS = [0, 5, 6.25, 10, 12.5, 15, 20, 25, 50, 100]  # 0 = auto
RFSQUELCH = ["OFF", "S1", "S2", "S3", "S4", "S5", "S6", "S7", "S8"]

# Charset is subset of ASCII + some unknown chars \x80-\x86
VALID_CHARS = ["%i" % int(x) for x in range(0, 10)] + \
    list(":>=<?@") + \
    [chr(x) for x in range(ord("A"), ord("Z") + 1)] + \
    list("[\\]_") + \
    [chr(x) for x in range(ord("a"), ord("z") + 1)] + \
    list("%*+,-/=$ ")

MEM_FORMAT = """
#seekto 0xceca;
struct {
  u8 unknown5;
  u8 unknown3;
  u8 unknown4:6,
     dsqtype:2;
  u8 dsqcode;
  u8 unknown1[2];
  char mycall[10];
  u8 unknown2[368];
} settings;

#seekto 0xfec9;
u8 checksum;
"""


@directory.register
class FTM3200Radio(ft1d.FT1Radio):
    """Yaesu FTM-3200D"""
    BAUD_RATE = 38400
    VENDOR = "Yaesu"
    MODEL = "FTM-3200D"
    VARIANT = "R"

    _model = b"AH52N"
    _memsize = 65227
    _block_lengths = [10, 65217]
    _has_vibrate = False
    _has_af_dual = False

    _mem_params = {
         "memnum": 199,            # size of memories array
         "flgnum": 199}            # size of flags array

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.pre_download = _(
            "1. Turn radio off.\n"
            "2. Connect cable to DATA terminal.\n"
            "3. Press and hold in the [MHz(SETUP)] key while turning the"
            " radio\n"
            "     on (\"CLONE\" will appear on the display).\n"
            "4. <b>After clicking OK</b>, press the [REV(DW)] key\n"
            "     to send image.\n")
        rp.pre_upload = _(
            "1. Turn radio off.\n"
            "2. Connect cable to DATA terminal.\n"
            "3. Press and hold in the [MHz(SETUP)] key while turning the"
            " radio\n"
            "     on (\"CLONE\" will appear on the display).\n"
            "4. Press the [MHz(SETUP)] key\n"
            "     (\"-WAIT-\" will appear on the LCD).\n")
        return rp

    def process_mmap(self):
        mem_format = ft1d.MEM_FORMAT + MEM_FORMAT
        self._memobj = bitwise.parse(mem_format % self._mem_params, self._mmap)

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_dtcs_polarity = False
        rf.valid_modes = list(MODES)
        rf.valid_tmodes = [x for x in TMODES if x is not None]
        rf.valid_cross_modes = [x for x in CROSS_MODES if x is not None]
        rf.valid_duplexes = list(ft1d.DUPLEX)
        rf.valid_tuning_steps = list(STEPS)
        rf.valid_bands = [(136000000, 174000000)]
        # rf.valid_skips = SKIPS
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_characters = "".join(VALID_CHARS)
        rf.valid_name_length = 8
        rf.memory_bounds = (1, 199)
        rf.can_odd_split = True
        rf.has_ctone = False
        rf.has_bank = False
        rf.has_bank_names = False
        # disable until implemented
        rf.has_settings = False
        return rf

    def _decode_label(self, mem):
        # TODO preserve the unknown \x80-x86 chars?
        return str(mem.label).rstrip("\xFF").rstrip()

    def _encode_label(self, mem):
        label = mem.name.rstrip().encode('ascii', 'ignore')
        return self._add_ff_pad(label, 16)

    def _encode_charsetbits(self, mem):
        # TODO this is a setting to decide if the memory should be displayed
        # as a name or frequency. Should we expose this setting to the user
        # instead of autoselecting it (and losing their preference)?
        if mem.name.rstrip() == '':
            return 0x0000
        return 0x0080

    def _decode_power_level(self, mem):
        return POWER_LEVELS[mem.power - 1]

    def _encode_power_level(self, mem):
        if mem.power is None:
            return 1
        else:
            return POWER_LEVELS.index(mem.power) + 1

    def _decode_mode(self, mem):
        return MODES[mem.mode_alt]

    def _encode_mode(self, mem):
        return MODES.index(mem.mode)

    def _get_tmode(self, mem, _mem):
        if _mem.tone_mode > 8:
            tmode = "Cross"
            mem.cross_mode = CROSS_MODES[_mem.tone_mode - 8]
        else:
            tmode = TMODES[_mem.tone_mode]

        if tmode == "Pager":
            # TODO chirp_common does not allow 'Pager'
            #   Expose as a different setting?
            mem.tmode = ""
        else:
            mem.tmode = tmode

    def _set_tmode(self, _mem, mem):
        if mem.tmode == "Cross":
            _mem.tone_mode = 8 + CROSS_MODES.index(mem.cross_mode)
        else:
            _mem.tone_mode = TMODES.index(mem.tmode)

    def _set_mode(self, _mem, mem):
        _mem.mode_alt = self._encode_mode(mem)

    def get_bank_model(self):
        return None

    def _debank(self, mem):
        return

    def _checksums(self):
        return [yaesu_clone.YaesuChecksum(0x064A, 0x06C8),
                yaesu_clone.YaesuChecksum(0x06CA, 0x0748),
                yaesu_clone.YaesuChecksum(0x074A, 0x07C8),
                yaesu_clone.YaesuChecksum(0x07CA, 0x0848),
                yaesu_clone.YaesuChecksum(0x0000, 0xFEC9)]

    def _get_settings(self):
        # TODO
        top = RadioSettings()
        return top

    @classmethod
    def _wipe_memory(cls, mem):
        mem.set_raw("\x00" * (mem.size() // 8))

    def sync_out(self):
        # Need to give enough time for the radio to ACK after writes
        self.pipe.timeout = 1
        return super(FTM3200Radio, self).sync_out()



# Copyright 2010 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from chirp.drivers import icf
from chirp import chirp_common, directory, bitwise
from chirp.chirp_common import to_GHz, from_GHz
from chirp.settings import RadioSetting, RadioSettingGroup, \
                RadioSettingValueBoolean, RadioSettingValueList, \
                RadioSettingValueInteger, RadioSettings


LOG = logging.getLogger(__name__)

MEM_FORMAT = """
struct {
  bbcd freq[3];
  u8  fractional:1,
      unknown:7;
  bbcd offset[2];
  u16 ctone:6,
      rtone:6,
      tune_step:4;
} memory[200];

#seekto 0x0690;
struct {
  u8 tmode:2,
     duplex:2,
     skip:1,
     pskip:1,
     mode:2;
} flags[200];

#seekto 0x0690;
u8 flags_whole[200];

#seekto 0x0767;
struct {
i8 rit;
u8 squelch;
u8 lock:1,
   ritfunct:1,
   unknown:6;
u8 unknown1[6];
u8 d_sel;
u8 autorp;
u8 priority;
u8 resume;
u8 pause;
u8 p_scan;
u8 bnk_scan;
u8 expand;
u8 ch;
u8 beep;
u8 light;
u8 ap_off;
u8 p_save;
u8 monitor;
u8 speed;
u8 edge;
u8 lockgroup;
} settings;

"""

TMODES = ["", "", "Tone", "TSQL", "TSQL"]  # last one is pocket beep
DUPLEX = ["", "", "-", "+"]
MODES = ["FM", "WFM", "AM", "Auto"]
STEPS = [5.0, 6.25, 10.0, 12.5, 15.0, 20.0, 25.0, 30.0, 50.0, 100.0]
AUTORP_LIST = ["Off", "Duplex Only", "Duplex and Tone"]
LOCKGROUP_LIST = ["Normal", "No Squelch", "No Volume", "All"]
SQUELCH_LIST = ["Open", "Auto"] + ["L%s" % x for x in range(1, 10)]
MONITOR_LIST = ["Push", "Hold"]
LIGHT_LIST = ["Off", "On", "Auto"]
PRIORITY_LIST = ["Off", "On", "Bell"]
BANKSCAN_LIST = ["Off", "Bank 0", "Bank 1"]
EDGE_LIST = ["%sP" % x for x in range(0, 20)] + ["Band", "All"]
PAUSE_LIST = ["%s sec" % x for x in range(2, 22, 2)] + ["Hold"]
RESUME_LIST = ["%s sec" % x for x in range(0, 6)]
APOFF_LIST = ["Off"] + ["%s min" % x for x in range(30, 150, 30)]
D_SEL_LIST = ["100 kHz", "1 MHz", "10 MHz"]


@directory.register
class ICQ7Radio(icf.IcomCloneModeRadio):
    """Icom IC-Q7A"""
    VENDOR = "Icom"
    MODEL = "IC-Q7A"

    _model = "\x19\x95\x00\x01"
    _memsize = 0x7C0
    _endframe = "Icom Inc\x2e"

    _ranges = [(0x0000, 0x07C0, 16)]

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.memory_bounds = (0, 199)
        rf.valid_modes = list(MODES)
        rf.valid_tmodes = list(TMODES)
        rf.valid_duplexes = list(DUPLEX)
        rf.valid_tuning_steps = list(STEPS)
        rf.valid_bands = [(1000000,   823995000),
                          (849000000, 868995000),
                          (894000000, 1309995000)]
        rf.valid_skips = ["", "S", "P"]
        rf.has_dtcs = False
        rf.has_dtcs_polarity = False
        rf.has_bank = False
        rf.has_name = False
        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_raw_memory(self, number):
        return (repr(self._memobj.memory[number]) +
                repr(self._memobj.flags[number]))

    def validate_memory(self, mem):
        if mem.freq < 30000000 and mem.mode != 'AM':
            return [chirp_common.ValidationError(
                'Only AM is allowed below 30 MHz')]
        return icf.IcomCloneModeRadio.validate_memory(self, mem)

    def get_memory(self, number):
        _mem = self._memobj.memory[number]
        _flag = self._memobj.flags[number]

        mem = chirp_common.Memory()
        mem.number = number
        if self._memobj.flags_whole[number] == 0xFF:
            mem.empty = True
            return mem

        mem.freq = int(_mem.freq) * 1000
        if _mem.fractional:
            mem.freq = chirp_common.fix_rounded_step(mem.freq)
        mem.offset = int(_mem.offset) * 1000
        try:
            mem.rtone = chirp_common.TONES[_mem.rtone]
        except IndexError:
            mem.rtone = 88.5
        try:
            mem.ctone = chirp_common.TONES[_mem.ctone]
        except IndexError:
            mem.ctone = 88.5
        try:
            mem.tuning_step = STEPS[_mem.tune_step]
        except IndexError:
            LOG.error("Invalid tune step index %i" % _mem.tune_step)
        mem.tmode = TMODES[_flag.tmode]
        mem.duplex = DUPLEX[_flag.duplex]
        if mem.freq < 30000000:
            mem.mode = "AM"
        else:
            mem.mode = MODES[_flag.mode]
        if _flag.pskip:
            mem.skip = "P"
        elif _flag.skip:
            mem.skip = "S"

        return mem

    def set_memory(self, mem):
        _mem = self._memobj.memory[mem.number]
        _flag = self._memobj.flags[mem.number]

        if mem.empty:
            self._memobj.flags_whole[mem.number] = 0xFF
            return

        _mem.set_raw("\x00" * 8)

        if mem.freq > to_GHz(1):
            _mem.freq = (mem.freq // 1000) - to_GHz(1)
            upper = from_GHz(mem.freq) << 4
            _mem.freq[0].clr_bits(0xF0)
            _mem.freq[0].set_bits(upper)
        else:
            _mem.freq = mem.freq / 1000
        _mem.fractional = chirp_common.is_fractional_step(mem.freq)
        _mem.offset = mem.offset / 1000
        _mem.rtone = chirp_common.TONES.index(mem.rtone)
        _mem.ctone = chirp_common.TONES.index(mem.ctone)
        _mem.tune_step = STEPS.index(mem.tuning_step)
        _flag.tmode = TMODES.index(mem.tmode)
        _flag.duplex = DUPLEX.index(mem.duplex)
        _flag.mode = MODES.index(mem.mode)
        _flag.skip = mem.skip == "S" and 1 or 0
        _flag.pskip = mem.skip == "P" and 1 or 0

    def get_settings(self):
        _settings = self._memobj.settings
        basic = RadioSettingGroup("basic", "Basic Settings")
        group = RadioSettings(basic)

        rs = RadioSetting("ch", "Channel Indication Mode",
                          RadioSettingValueBoolean(_settings.ch))
        basic.append(rs)

        rs = RadioSetting("expand", "Expanded Settings Mode",
                          RadioSettingValueBoolean(_settings.expand))
        basic.append(rs)

        rs = RadioSetting("beep", "Beep Tones",
                          RadioSettingValueBoolean(_settings.beep))
        basic.append(rs)

        rs = RadioSetting("autorp", "Auto Repeater Function",
                          RadioSettingValueList(
                              AUTORP_LIST, AUTORP_LIST[_settings.autorp]))
        basic.append(rs)

        rs = RadioSetting("ritfunct", "RIT Runction",
                          RadioSettingValueBoolean(_settings.ritfunct))
        basic.append(rs)

        rs = RadioSetting("rit", "RIT Shift (kHz)",
                          RadioSettingValueInteger(-7, 7, _settings.rit))
        basic.append(rs)

        rs = RadioSetting("lock", "Lock",
                          RadioSettingValueBoolean(_settings.lock))
        basic.append(rs)

        rs = RadioSetting("lockgroup", "Lock Group",
                          RadioSettingValueList(
                              LOCKGROUP_LIST,
                              LOCKGROUP_LIST[_settings.lockgroup]))
        basic.append(rs)

        rs = RadioSetting("squelch", "Squelch",
                          RadioSettingValueList(
                              SQUELCH_LIST, SQUELCH_LIST[_settings.squelch]))
        basic.append(rs)

        rs = RadioSetting("monitor", "Monitor Switch Function",
                          RadioSettingValueList(
                              MONITOR_LIST,
                              MONITOR_LIST[_settings.monitor]))
        basic.append(rs)

        rs = RadioSetting("light", "Display Backlighting",
                          RadioSettingValueList(
                              LIGHT_LIST, LIGHT_LIST[_settings.light]))
        basic.append(rs)

        rs = RadioSetting("priority", "Priority Watch Operation",
                          RadioSettingValueList(
                              PRIORITY_LIST,
                              PRIORITY_LIST[_settings.priority]))
        basic.append(rs)

        rs = RadioSetting("p_scan", "Frequency Skip Function",
                          RadioSettingValueBoolean(_settings.p_scan))
        basic.append(rs)

        rs = RadioSetting("bnk_scan", "Memory Bank Scan Selection",
                          RadioSettingValueList(
                              BANKSCAN_LIST,
                              BANKSCAN_LIST[_settings.bnk_scan]))
        basic.append(rs)

        rs = RadioSetting("edge", "Band Edge Scan Selection",
                          RadioSettingValueList(
                              EDGE_LIST, EDGE_LIST[_settings.edge]))
        basic.append(rs)

        rs = RadioSetting("pause", "Scan Pause Time",
                          RadioSettingValueList(
                              PAUSE_LIST, PAUSE_LIST[_settings.pause]))
        basic.append(rs)

        rs = RadioSetting("resume", "Scan Resume Time",
                          RadioSettingValueList(
                              RESUME_LIST, RESUME_LIST[_settings.resume]))
        basic.append(rs)

        rs = RadioSetting("p_save", "Power Saver",
                          RadioSettingValueBoolean(_settings.p_save))
        basic.append(rs)

        rs = RadioSetting("ap_off", "Auto Power-off Function",
                          RadioSettingValueList(
                              APOFF_LIST, APOFF_LIST[_settings.ap_off]))
        basic.append(rs)

        rs = RadioSetting("speed", "Dial Speed Acceleration",
                          RadioSettingValueBoolean(_settings.speed))
        basic.append(rs)

        rs = RadioSetting("d_sel", "Dial Select Step",
                          RadioSettingValueList(
                              D_SEL_LIST, D_SEL_LIST[_settings.d_sel]))
        basic.append(rs)

        return group

    def set_settings(self, settings):
        _settings = self._memobj.settings
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    name = element.get_name()
                    if "." in name:
                        bits = name.split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            if "/" in bit:
                                bit, index = bit.split("/", 1)
                                index = int(index)
                                obj = getattr(obj, bit)[index]
                            else:
                                obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = _settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    else:
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise



# Copyright 2019 Dan Clemmensen <DanClemmensen@Gmail.com>
#    Derives loosely from two sources released under GPLv2:
#      ./template.py, Copyright 2012 Dan Smith <dsmith@danplanet.com>
#      ./ft60.py, Copyright 2011 Dan Smith <dsmith@danplanet.com>
#    Edited 2020, 2023 Bernhard Hailer AE6YN <ham73tux@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
CHIRP driver for Yaesu radios that use the SCU-35 cable. This includes at
least the FT-4X, FT-4V, FT-65, and FT-25. This driver will not work with older
Yaesu models.
"""
import logging
import struct
import copy
from chirp import chirp_common, directory, memmap, bitwise, errors, util
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueList, RadioSettingValueString, RadioSettings

LOG = logging.getLogger(__name__)


# Layout of Radio memory image.
# This module and the serial protocol treat the FT-4 memory as 16-byte blocks.
# There in nothing magic about 16-byte blocks, but it simplifies the
# description. There are 17 groups of blocks, each with a different purpose
# and format. Five groups consist of channel memories, or "mems" in CHIRP.
# A "mem" describes a radio channel, and all "mems" have the same internal
# format. Three of the groups consist of bitmaps, which all have the same
# internal mapping. also groups for Name, misc, DTMF digits, and prog,
# plus some unused groups. The MEM_FORMAT describes the radio image memory.
# MEM_FORMAT is parsed in module ../bitwise.py. Syntax is similar but not
# identical to C data and structure definitions.

# Define the structures for each type of group here, but do not associate them
# with actual memory addresses yet
MEM_FORMAT = """
struct slot {
 u8 tx_pwr;     //0, 1, 2 == lo, medium, high
 bbcd freq[4];  // Hz/10   but must end in 00
 u8 tx_ctcss;   //see ctcss table, but radio code = CHIRP code+1. 0==off
 u8 rx_ctcss;   //see ctcss table, but radio code = CHIRP code+1. 0==off
 u8 tx_dcs;     //see dcs table, but radio code = CHIRP code+1. 0==off
 u8 rx_dcs;     //see dcs table, but radio code = CHIRP code+1. 0==off
 u8 unknown1:5,
    duplex:3;   //(auto,offset). (0,2,4,5)= (+,-,0, auto)
 ul16 offset;   //little-endian binary * scaler, +- per duplex
                   //scaler is 25 kHz for FT-4, 50 kHz for FT-65.
 u8 unknown2:7,
    tx_width:1;  //0=wide, 1=narrow
 u8 step;       //STEPS (0-9)=(auto,5,6.25,10,12.5,15,20,25,50,100) kHz
 u8 sql_type;   //(0-6)==(off,r-tone,t-tone,tsql,rev tn,dcs,pager)
 u8 unused;
};
// one bit per channel. 220 bits (200 mem+ 2*10 PMS) padded to fill
//exactly 2 blocks
struct bitmap {
u8 b8[28];
u8 unused[4];
};
//name struct occupies half a block (8 bytes)
//the code restricts the actual len to 6 for an FT-4
struct name {
  u8 chrs[8];    //[0-9,A-z,a-z, -] padded with spaces
};
//txfreq struct occupies 4 bytes (1/4 slot)
struct txfreq {
 bbcd freq[4];
};

//miscellaneous params. One 4-block group.
//"SMI": "Set Mode Index" of the FT-4 radio keypad function to set parameter.
//"SMI numbers on the FT-65 are different but the names in mem are the same.
struct misc {
  u8  apo;        //SMI 01. 0==off, (1-24) is the number of half-hours.
  u8  arts_beep;  //SMI 02. 0==off, 1==inrange, 2==always
  u8  arts_intv;  //SMI 03. 0==25 seconds, 1==15 seconds
  u8  battsave;   //SMI 32. 0==off, (1-5)==(200,300,500,1000,2000) ms
  u8  bclo;       //SMI 04. 0==off, 1==on
  u8  beep;       //SMI 05. (0-2)==(key+scan,key, off)
  u8  bell;       //SMI 06. (0-5)==(0,1,3,5,8,continuous) bells
  u8  cw_id[6];   //SMI 08. callsign (A_Z,0-9) (pad with space if <6)
  u8  unknown1[3];
  // addr= 2010
  u8  dtmf_mode;  //SMI 12. 0==manual, 1==auto
  u8  dtmf_delay; //SMI 11. (0-4)==(50,250,450,750,1000) ms
  u8  dtmf_speed; //SMI 13. (0,1)=(50,100) ms
  u8  edg_beep;   //SMI 14. 0==off, 1==on
  u8  key_lock;   //SMI 18. (0-2)==(key,ptt,key+ptt)
  u8  lamp;       //SMI 15. (0-4)==(5,10,30,continuous,off) secKEY
  u8  tx_led;     //SMI 17. 0==off, 1==on
  u8  bsy_led;    //SMI 16. 0==off, 1==on
  u8  moni_tcall; //SMI 19. (0-4)==(mon,1750,2100,1000,1450) tcall Hz.
  u8  pri_rvt;    //SMI 23. 0==off, 1==on
  u8  scan_resume; //SMI 34. (0-2)==(busy,hold,time)
  u8  rf_squelch;  //SMI 28. 0==off, 8==full, (1-7)==(S1-S7)
  u8  scan_lamp;  //SMI 33  0==off,1==on
  u8  unknown2;
  u8  use_cwid;   //SMI 7. 0==no, 1==yes
  u8  compander;  // compander on FT_65
  // addr 2020
  u8  unknown3;
  u8  tx_save;    //SMI 41. 0==off, 1==on (addr==2021)
  u8  vfo_spl;    //SMI 42. 0==off, 1==on
  u8  vox;        //SMI 43. 0==off, 1==on
  u8  wfm_rcv;    //SMI 44. 0==off, 1==on
  u8  unknown4;
  u8  wx_alert;   //SMI 46. 0==off, 1==0n
  u8  tot;        //SMI 39. 0-off, (1-30)== (1-30) minutes
  u8  pager_tx1;  //SMI 21. (0-49)=(1-50) epcs code (i.e., value is code-1)
  u8  pager_tx2;  //SMI 21   same
  u8  pager_rx1;  //SMI 23   same
  u8  pager_rx2;  //SMI 23   same
  u8  pager_ack;  //SMI 22   same
  u8  unknown5[3];  //possibly sql_setting and pgm_vfo_scan on FT-65?
  // addr 2030
  u8  use_passwd; //SMI 26 0==no, 1==yes
  u8  passwd[4];  //SMI 27 ASCII (0-9)
  u8  unused2[11]; //  pad out to a block boundary
};

struct dtmfset {
 u8 digit[16];    //ASCII (*,#,-,0-9,A-D). (dash-filled)
};

// area to be filled with 0xff, or just ignored
struct notused {
  u8 unused[16];
};
// areas we are still analyzing
struct unknown {
  u8 notknown[16];
};

struct progc {
  u8 usage;          //P key is (0-2) == unused, use P channel, use parm
  u8 submode:2,      //if parm!=0 submode 0-3 of mode
     parm:6;         //if usage == 2: if 0, use m-channel, else  mode
};
"""
# Actual memory layout. 0x215 blocks, in 20 groups.
MEM_FORMAT += """
#seekto 0x0000;
struct unknown radiotype;     //0000 probably a radio type ID but not sure
struct slot    memory[200];   //0010 channel memory array
struct slot    pms[20];       //0c90 10 PMS (L,U) slot pairs
struct slot    vfo[5];        //0dd0 VFO (A UHF, A VHF, B FM, B  UHF, B VHF)
struct slot    home[3];       //0e20 Home (FM, VHF, UHF)
struct bitmap  enable;        //0e50
struct bitmap  scan;          //0e70
struct notused notused0;      //0e90
struct bitmap  bankmask[10];  //0ea0
struct notused notused1[2];   //0fe0
struct name  names[220];      //1000 220 names in 110 blocks
struct notused notused2[2];   //16e0
struct txfreq  txfreqs[220];  //1700 220 freqs in 55 blocks
struct notused notused3[89];  //1a20
struct misc  settings;        //2000  4-block collection of misc params
struct notused notused4[2];   //2040
struct dtmfset dtmf[9];       //2060  sets 1-9
struct notused notused5;      //20f0
//struct progs progkeys;      //2100  not a struct. bitwise.py refuses
struct progc  progctl[4];        //2100 8 bytes. 1 2-byte struct per P key
u8 pmemndx[4];                   //2108 4 bytes, 1 per P key
u8 notused6[4];                  //210c fill out the block
struct slot  prog[4];         //2110  P key "channel" array
//---------------- end of FT-4 mem?
"""
# The remaining mem is (apparently) not available on the FT4 but is
# reported to be available on the FT-65. Not implemented here yet.
# Possibly, memory-mapped control registers that  allow for "live-mode"
# operation instead of "clone-mode" operation.
# 2150 27ff                   (unused?)
# 2800 285f       6           MRU operation?
# 2860 2fff                   (unused?)
# 3000 310f       17          (version info, etc?)
# ----------END of memory map


# Begin Serial transfer utilities for the SCU-35 cable.

# The serial transfer protocol was implemented here after snooping the wire.
# After it was implemented, we noticed that it is identical to the protocol
# implemented in th9000.py. A non-echo version is implemented in anytone_ht.py.
#
# The pipe.read and pipe.write functions use bytes, not strings. The serial
# transfer utilities operate only to move data between the memory object and
# the serial port. The code runs on either Python 2 or Python3, so some
# constructs could be better optimized for one or the other, but not both.


def get_mmap_data(radio):
    """
    horrible kludge needed until we convert entirely to Python 3 OR we add a
    slightly less horrible kludge to the Py2 or Py3 versions of memmap.py.
    The minimal change have Py3 code return a bytestring instead of a string.
    This is the only function in this module that must explicitly test for the
    data string type. It is used only in the do_upload function.
    returns the memobj data as a byte-like object.
    """
    data = radio.get_mmap().get_packed()
    if isinstance(data, bytes):
        return data
    return bytearray(radio.get_mmap()._data)


def checkSum8(data):
    """
    Calculate the 8 bit checksum of buffer
    Input: buffer   - bytes
    returns: integer
    """
    return sum(x for x in bytearray(data)) & 0xFF


def variable_len_resp(pipe):
    """
    when length of expected reply is not known, read byte at a time
    until the ack character is found.
    """
    response = b""
    i = 0
    toolong = 256        # arbitrary
    while True:
        b = pipe.read(1)
        if b == b'\x06':
            break
        else:
            response += b
            i += 1
            if i > toolong:
                LOG.debug("Response too long. got" + util.hexprint(response))
                raise errors.RadioError("Response from radio too long.")
    return response


def sendcmd(pipe, cmd, response_len, retry=0):
    """
    send a command bytelist to radio,receive and return the resulting bytelist.
    Input: pipe         - serial port object to use
           cmd          - bytes to send
           response_len - number of bytes of expected response,
                          not including the ACK. (if None, read until ack)
    This cable is "two-wire": The TxD and RxD are "or'ed" so we receive
    whatever we send and then whatever response the radio sends. We check the
    echo and strip it, returning only the radio's response.
    We also check and strip the ACK character at the end of the response.
    """
    pipe.write(cmd)
    echo = pipe.read(len(cmd))
    if echo != cmd:
        msg = "Bad echo. Sent:" + util.hexprint(cmd) + ", "
        msg += "Received:" + util.hexprint(echo)
        LOG.debug(msg)
        raise errors.RadioError(
            "Incorrect echo on serial port. Radio off? Bad cable?")
    if response_len is None:
        return variable_len_resp(pipe)
    if response_len > 0:
        response = pipe.read(response_len)
    else:
        response = b""
    ack = pipe.read(1)
    if ack != b'\x06':
        if retry < 5:
            LOG.debug("retry: " + str(retry))
            retry += 1
            return sendcmd(pipe, cmd, response_len, retry)
        else:
            LOG.debug("missing ack: expected 0x06, got" + util.hexprint(ack))
            raise errors.RadioError("Incorrect ACK on serial port.")
    return response


def enter_clonemode(radio):
    """
    Send the PROGRAM command and check the response. Retry if
    needed. After 3 tries, send an "END" and try some more if
    it is acknowledged.
    """
    for use_end in range(0, 3):
        for i in range(0, 3):
            try:
                if b"QX" == sendcmd(radio.pipe, b"PROGRAM", 2):
                    return
            except Exception:
                continue
        sendcmd(radio.pipe, b"END", 0)
    raise errors.RadioError("expected QX from radio.")


def startcomms(radio, way):
    """
    For either upload or download, put the radio into PROGRAM mode
    and check the radio's ID. In this preliminary version of the driver,
    the exact nature of the ID has been inferred from a single test case.
        set up the progress bar
        send "PROGRAM" to command the radio into clone mode
        read the initial string (version?)
    """
    progressbar = chirp_common.Status()
    progressbar.msg = "Cloning " + way + " radio"
    progressbar.max = radio.numblocks
    enter_clonemode(radio)
    id_response = sendcmd(radio.pipe, b'\x02', None)

    # The last byte of the ID contains info about regional differences
    radio.subtype = id_response[-1]
    # Set last byte of the ID zero so that no ID warning is thrown
    # Unfortunately, the returned object is an immutable bytes object,
    # so we need to convert into a bytearray, modify, and convert back.
    ba_id_response = bytearray(id_response)
    ba_id_response[len(ba_id_response) - 1] = 0
    id_response_mod = bytes(ba_id_response)

    if id_response != radio.id_str:
        substr0 = radio.id_str[:radio.id_str.find(b'\x00')]
        if id_response[:id_response.find(b'\x00')] != substr0:
            msg = "ID mismatch. Expected" + util.hexprint(radio.id_str)
            msg += ", Received:" + util.hexprint(id_response_mod)
            LOG.warning(msg)
            raise errors.RadioError("Incorrect ID read from radio.")
        else:
            msg = "ID suspect. Expected" + util.hexprint(radio.id_str)
            msg += ", Received:" + util.hexprint(id_response_mod)
            LOG.warning(msg)
    return progressbar


def getblock(pipe, addr, image):
    """
    read a single 16-byte block from the radio.
    send the command and check the response
    places the response into the correct offset in the supplied bytearray
    returns True if successful, False if error.
    """
    cmd = struct.pack(">cHb", b"R", addr, 16)
    response = sendcmd(pipe, cmd, 21)
    if (response[0] != b"W"[0]) or (response[1:4] != cmd[1:4]):
        msg = "Bad response. Sent:\n%s" % util.hexprint(cmd)
        msg += "\nReceived:\n%s" % util.hexprint(response)
        LOG.debug(msg)
        return False
    if checkSum8(response[1:20]) != bytearray(response)[20]:
        LOG.debug("Bad checksum:\n%s" % util.hexprint(response))
        LOG.debug('%r != %r' % (checkSum8(response[1:20]),
                                bytearray(response)[20]))
        return False
    image[addr:addr+16] = response[4:20]
    return True


def do_download(radio):
    """
    Read memory from the radio.
      call startcomms to go into program mode and check version
      create an mmap
      read the memory blocks and place the data into the mmap
      send "END"
    """
    image = bytearray(radio.get_memsize())
    pipe = radio.pipe  # Get the serial port connection
    progressbar = startcomms(radio, "from")
    for blocknum in range(radio.numblocks):
        for i in range(0, 3):
            if getblock(pipe, 16 * blocknum, image):
                break
            if i == 2:
                raise errors.RadioError(
                   "read block from radio failed 3 times")
        progressbar.cur = blocknum
        radio.status_fn(progressbar)
    sendcmd(pipe, b"END", 0)
    return memmap.MemoryMapBytes(bytes(image))


def putblock(pipe, addr, data):
    """
    write a single 16-byte block to the radio
    send the command and check the response
    """
    chkstr = struct.pack(">Hb",  addr, 16) + data
    msg = b'W' + chkstr + struct.pack('B', checkSum8(chkstr)) + b'\x06'
    sendcmd(pipe, msg, 0)


def do_upload(radio):
    """
    Write memory image to radio
      call startcomms to go into program mode and check version
      write the memory blocks. Skip the first block
      send "END"
    """
    pipe = radio.pipe  # Get the serial port connection
    progressbar = startcomms(radio, "to")
    data = get_mmap_data(radio)
    for _i in range(1, radio.numblocks):
        putblock(pipe, 16*_i, data[16*_i:16*(_i+1)])
        progressbar.cur = _i
        radio.status_fn(progressbar)
    sendcmd(pipe, b"END", 0)
    return
# End serial transfer utilities


def bit_loc(bitnum):
    """
    return the ndx and mask for a bit location
    """
    return (bitnum // 8, 1 << (bitnum & 7))


def store_bit(bankmem, bitnum, val):
    """
    store a bit in a bankmem. Store 0 or 1 for False or True
    """
    ndx, mask = bit_loc(bitnum)
    if val:
        bankmem.b8[ndx] |= mask
    else:
        bankmem.b8[ndx] &= ~mask
    return


def retrieve_bit(bankmem, bitnum):
    """
    return True or False for a bit in a bankmem
    """
    ndx, mask = bit_loc(bitnum)
    return (bankmem.b8[ndx] & mask) != 0


# A bank is a bitmap of 220 bits. 200 mem slots and 2*10 PMS slots.
# There are 10 banks.
class YaesuSC35GenericBankModel(chirp_common.BankModel):

    def get_num_mappings(self):
        return 10

    def get_mappings(self):
        banks = []
        for i in range(1, 1 + self.get_num_mappings()):
            bank = chirp_common.Bank(self, "%i" % i, "Bank %i" % i)
            bank.index = i - 1
            banks.append(bank)
        return banks

    def add_memory_to_mapping(self, memory, bank):
        bankmem = self._radio._memobj.bankmask[bank.index]
        store_bit(bankmem, memory.number-1, True)

    def remove_memory_from_mapping(self, memory, bank):
        bankmem = self._radio._memobj.bankmask[bank.index]
        if not retrieve_bit(bankmem, memory.number-1):
            raise Exception("Memory %i is not in bank %s." %
                            (memory.number, bank))
        store_bit(bankmem, memory.number-1, False)

    # return a list of slots in a bank
    def get_mapping_memories(self, bank):
        memories = []
        for i in range(*self._radio.get_features().memory_bounds):
            if retrieve_bit(self._radio._memobj.bankmask[bank.index], i - 1):
                memories.append(self._radio.get_memory(i))
        return memories

    # return a list of banks a slot is a member of
    def get_memory_mappings(self, memory):
        memndx = memory.number - 1
        banks = []
        for bank in self.get_mappings():
            if retrieve_bit(self._radio._memobj.bankmask[bank.index], memndx):
                banks.append(bank)
        return banks


# the values in these lists must also be in the canonical UI list
# we can re-arrange the order, and we don't need to have all
# the values, but we cannot add our own values here.

DUPLEX = ["+", "", "-", "", "off", "", "split"]  # (0,2,4,5) = (+, -, 0, auto)
# The radio implements duplex "auto" as 5. We map to "". It is a convenience
# function in the radio that affects the offset, which sets the duplex value
# according to the frequency in use. Yaesu doesn't entirely adhere to the band
# plans; but worse, they save the value 'auto' instead of + or -. Why Yaesu
# is doing such a thing is beyond me. [AE6YN]
DUPLEX_AUTO_US = [
    [145100000, 145495000, 2],
    [146000000, 146395000, 0],
    [146600000, 146995000, 2],
    [147000000, 147395000, 0],
    [147600000, 147995000, 2]]
# (There are no automatic duplex values in IARU-2 70CM.)
DUPLEX_AUTO_EU = [
    [145600000, 145800000, 2],
    [438200000, 439425000, 2]]

SKIPS = ["S", ""]

POWER_LEVELS = [
    chirp_common.PowerLevel("High", watts=5.0),  # high must be first (0)
    chirp_common.PowerLevel("Mid", watts=2.5),
    chirp_common.PowerLevel("Low", watts=0.5)]

# these steps encode to 0-9 on all radios, but encoding #2 is disallowed
# on the US versions (FT-4XR)
STEP_CODE = [0, 5.0, 6.25, 10.0, 12.5, 15.0, 20.0, 25.0, 50.0, 100.0]
US_LEGAL_STEPS = list(STEP_CODE)  # copy to pass to UI on US radios
US_LEGAL_STEPS.remove(6.25)       # euro radios just use STEP_CODE

# Map the radio image sql_type (0-6) to the CHIRP mem values.
# Yaesu "TSQL" and "DCS" each map to different CHIRP values depending on the
# radio values of the tx and rx tone codes. The table is a list of rows, one
# per Yaesu sql_type (0-5). The code does not use this table when the sql_type
# is 6 (PAGER). Each row is a tuple. Its first member is a list of
# [tmode,cross] or [tmode, cross, suppress]. "Suppress" is used only when
# encoding UI-->radio. When decoding radio-->UI, two of the sql_types each
# result in 5 possible UI decodings depending on the tx and rx codes, and the
# list in each of these rows has five members. These two row tuples each have
# two additional members to specify which of the radio fields to examine.
# The map from CHIRP UI to radio image types is also built from this table.
RADIO_TMODES = [
        ([["", None], ], ),            # sql_type= 0. off
        ([["Cross", "->Tone"], ], ),   # sql_type= 1. R-TONE
        ([["Tone", None], ], ),        # sql_type= 2. T-TONE
        ([                             # sql_type= 3. TSQL:
          ["", None],                       # tx==0, rx==0 : invalid
          ["TSQL", None],                   # tx==0
          ["Tone", None],                   # rx==0
          ["Cross", "Tone->Tone"],          # tx!=rx
          ["TSQL", None]                    # tx==rx
         ], "tx_ctcss", "rx_ctcss"),     # tx and rx fields to check
        ([["TSQL-R", None], ], ),      # sql_type= 4. REV TN
        ([                             # sql_type= 5.DCS:
          ["", None],                       # tx==0, rx==0 : invalid
          ["Cross", "->DTCS", "tx_dcs"],    # tx==0. suppress tx
          ["Cross", "DTCS->", "rx_dcs"],    # rx==0. suppress rx
          ["Cross", "DTCS->DTCS"],          # tx!=rx
          ["DTCS", None]                    # tx==rx
         ], "tx_dcs", "rx_dcs"),         # tx and rx fields to check
        #                              # sql_type= 6. PAGER is a CHIRP "extra"
        ]

# Find all legal values for the tmode and cross fields for the UI.
# We build two dictionaries to do the lookups when encoding.
# The reversed range is a kludge: by happenstance, earlier duplicates
# in the above table are the preferred mapping, they override the
# later ones when we process the table backwards.
TONE_DICT = {}          # encode sql_type.
CROSS_DICT = {}         # encode sql_type.

for sql_type in reversed(range(0, len(RADIO_TMODES))):
    sql_type_row = RADIO_TMODES[sql_type]
    for decode_row in sql_type_row[0]:
        suppress = None
        if len(decode_row) == 3:
            suppress = decode_row[2]
        TONE_DICT[decode_row[0]] = (sql_type, suppress)
        if decode_row[1]:
            CROSS_DICT[decode_row[1]] = (sql_type, suppress)

# The keys are added to the "VALID" lists using code that puts those lists
# in the same order as the UI's default order instead of the random dict
# order or the arbitrary build order.
VALID_TONE_MODES = []   # list for UI.
VALID_CROSS_MODES = []  # list for UI.
for name in chirp_common.TONE_MODES:
    if name in TONE_DICT:
        VALID_TONE_MODES += [name]
for name in chirp_common.CROSS_MODES:
    if name in CROSS_DICT:
        VALID_CROSS_MODES += [name]


DTMF_CHARS = "0123456789ABCD*#- "
CW_ID_CHARS = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ "
PASSWD_CHARS = "0123456789"
CHARSET = CW_ID_CHARS + "abcdefghijklmnopqrstuvwxyz*+-/@"
PMSNAMES = ["%s%02d" % (c, i) for i in range(1, 11) for c in ('L', 'U')]

# Four separate arrays of special channel mems.
# Each special has unique constraints: band, name yes/no, and pms L/U
# The FT-65 class replaces the "prog" entry in this list.
# The name field must be the name of a slot array in MEM_FORMAT
SPECIALS_FT4 = [
    ("pms", PMSNAMES),
    ("vfo", ["VFO A UHF", "VFO A VHF", "VFO B FM", "VFO B VHF", "VFO B UHF"]),
    ("home", ["HOME FM", "HOME VHF", "HOME UHF"]),
    ("prog", ["P1", "P2"])
    ]
SPECIALS_FT65 = SPECIALS_FT4
FT65_PROGS = ("prog", ["P1", "P2", "P3", "P4"])
SPECIALS_FT65[-1] = FT65_PROGS    # replace the last entry (P key names)


VALID_BANDS_DUAL = [
    (65000000, 108000000),     # broadcast FM, receive only
    (136000000, 174000000),    # VHF
    (400000000, 480000000)     # UHF
    ]

VALID_BANDS_VHF = [
    (65000000, 108000000),     # broadcast FM, receive only
    (136000000, 174000000),    # VHF
    ]

# bands for the five VFO and three home channel memories
BAND_ASSIGNMENTS_DUALBAND = [2, 1, 0, 1, 2, 0, 1, 2]  # all locations used
BAND_ASSIGNMENTS_MONO_VHF = [1, 1, 0, 1, 1, 0, 1, 1]  # UHF locations unused


# None, and 50 Tones. Use this explicit array because the
# one in chirp_common could change and no longer describe our radio
TONE_MAP = [
    None, 67.0, 69.3, 71.9, 74.4, 77.0, 79.7, 82.5,
    85.4, 88.5, 91.5, 94.8, 97.4, 100.0, 103.5,
    107.2, 110.9, 114.8, 118.8, 123.0, 127.3,
    131.8, 136.5, 141.3, 146.2, 151.4, 156.7,
    159.8, 162.2, 165.5, 167.9, 171.3, 173.8,
    177.3, 179.9, 183.5, 186.2, 189.9, 192.8,
    196.6, 199.5, 203.5, 206.5, 210.7, 218.1,
    225.7, 229.1, 233.6, 241.8, 250.3, 254.1
    ]

# None, and 104 DTCS Codes. Use this explicit array because the
# one in chirp_common could change and no longer describe our radio
DTCS_MAP = [
    None, 23,  25,  26,  31,  32,  36,  43,  47,  51,  53,  54,
    65,  71,  72,  73,  74,  114, 115, 116, 122, 125, 131,
    132, 134, 143, 145, 152, 155, 156, 162, 165, 172, 174,
    205, 212, 223, 225, 226, 243, 244, 245, 246, 251, 252,
    255, 261, 263, 265, 266, 271, 274, 306, 311, 315, 325,
    331, 332, 343, 346, 351, 356, 364, 365, 371, 411, 412,
    413, 423, 431, 432, 445, 446, 452, 454, 455, 462, 464,
    465, 466, 503, 506, 516, 523, 526, 532, 546, 565, 606,
    612, 624, 627, 631, 632, 654, 662, 664, 703, 712, 723,
    731, 732, 734, 743, 754
    ]

# The legal PAGER codes are the same as the CTCSS codes, but we
# pass them to the UI as a list of strings
EPCS_CODES = [format(flt) for flt in [0] + TONE_MAP[1:]]


# allow a child class to add a param to its class
# description list. used when a specific radio class has
# a param that is not in the generic list.
def add_paramdesc(desc_list, group, param):
    for description in desc_list:
        groupname, title, parms = description
        if group == groupname:
            description[2].append(param)
            return


class YaesuSC35GenericRadio(chirp_common.CloneModeRadio,
                            chirp_common.ExperimentalRadio):
    """
    Base class for all Yaesu radios using the SCU-35 programming cable
    and its protocol. Classes for sub families extend this class and
    are found towards the end of this file.
    """
    VENDOR = "Yaesu"
    MODEL = "SCU-35Generic"  # No radio directly uses the base class
    BAUD_RATE = 9600
    MAX_MEM_SLOT = 200
    NEEDS_COMPAT_SERIAL = False

    # These settings are common to all radios in this family.
    _valid_chars = chirp_common.CHARSET_ASCII
    numblocks = 0x215           # number of 16-byte blocks in the radio
    _memsize = 16 * numblocks   # used by CHIRP file loader to guess radio type
    MAX_MEM_SLOT = 200

    # ID string last byte can contain extra info. It is usually 0;
    # however, on FT25R/65R sold in Asia we have seen that it can be 3,
    # which affects the scaler: instead of the default 50000 it's then 25000.
    @property
    def subtype(self):
        return self.metadata.get('ft4_subtype', 0)

    @subtype.setter
    def subtype(self, value):
        self.metadata = {'ft4_subtype': int(value)}

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = (
            'Tested and mostly works, but may give you issues\n'
            'when using lesser common radio variants.\n'
            'Proceed at your own risk, and let us know about issues!'
            )

        rp.pre_download = "".join([
            "1. Connect programming cable to MIC jack.\n",
            "2. Press OK."
            ]
            )
        rp.pre_upload = rp.pre_download
        return rp

    # identify the features that can be manipulated on this radio.
    # mentioned here only when different from defaults in chirp_common.py
    def get_features(self):

        rf = chirp_common.RadioFeatures()
        specials = [name for s in self.class_specials for name in s[1]]
        rf.valid_special_chans = specials
        rf.memory_bounds = (1, self.MAX_MEM_SLOT)
        rf.valid_duplexes = DUPLEX
        rf.valid_tmodes = VALID_TONE_MODES
        rf.valid_cross_modes = VALID_CROSS_MODES
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_tuning_steps = self.legal_steps
        rf.valid_skips = SKIPS
        rf.valid_characters = CHARSET
        rf.valid_name_length = self.namelen
        rf.valid_modes = ["FM", "NFM"]
        rf.valid_bands = self.valid_bands
        rf.can_odd_split = True
        rf.has_ctone = True
        rf.has_rx_dtcs = True
        rf.has_dtcs_polarity = False    # REV TN reverses the tone, not the dcs
        rf.has_cross = True
        rf.has_settings = True

        return rf

    def get_bank_model(self):
        return YaesuSC35GenericBankModel(self)

    # read and parse the radio memory
    def sync_in(self):
        try:
            self._mmap = do_download(self)
        except Exception as e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)
        self.process_mmap()

    # write the memory image to the radio
    def sync_out(self):
        try:
            do_upload(self)
        except Exception as e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    # There are about 40 settings and most are handled generically in
    # get_settings. get_settings invokes these handlers for the few
    # that are more complicated.

    # callback for setting  byte arrays: DTMF[0-9], passwd, and CW_ID
    def apply_str_to_bytearray(self, element, obj):
        lng = len(obj)
        string = (element.value.get_value() + "                ")[:lng]
        bytes = bytearray(string, "ascii")
        for x in range(0, lng):    # memobj cannot iterate, so byte-by-byte
            obj[x] = bytes[x]
        return

    # add a string value to the RadioSettings
    def get_string_setting(self, obj, valid_chars, desc1, desc2, group):
        content = ''
        maxlen = len(obj)
        for x in range(0, maxlen):
            content += chr(obj[x])
        val = RadioSettingValueString(0, maxlen, content, True, valid_chars)
        rs = RadioSetting(desc1, desc2, val)
        rs.set_apply_callback(self.apply_str_to_bytearray, obj)
        group.append(rs)

    # called when found in the group_descriptions table to handle string value
    def get_strset(self, group, parm):
        #   parm =(paramname, paramtitle,(handler,[handler params])).
        objname, title, fparms = parm
        myparms = fparms[1]
        obj = getattr(self._memobj.settings,  objname)
        self.get_string_setting(obj, myparms[0], objname, title, group)

    # called when found in the group_descriptions table for DTMF strings
    def get_dtmfs(self, group, parm):
        objname, title, fparms = parm
        for i in range(1, 10):
            dtmf_digits = self._memobj.dtmf[i - 1].digit
            self.get_string_setting(
                dtmf_digits, DTMF_CHARS,
                "dtmf_%i" % i, "DTMF Autodialer Memory %i" % i, group)

    def apply_P(self, element, pnum, memobj):
        memobj.progctl[pnum].usage = element.value

    def apply_Pmode(self, element, pnum, memobj):
        memobj.progctl[pnum].parm = element.value

    def apply_Psubmode(self, element, pnum, memobj):
        memobj.progctl[pnum].submode = element.value

    def apply_Pmem(self, element, pnum, memobj):
        memobj.pmemndx[pnum] = element.value

    MEMLIST = ["%d" % i for i in range(1, MAX_MEM_SLOT + 1)] + PMSNAMES
    USAGE_LIST = ["unused", "P channel", "mode or M channel"]

    # called when found in the group_descriptions table
    # returns the settings for the programmable keys (P1-P4)
    def get_progs(self, group, parm):
        _progctl = self._memobj.progctl
        _progndx = self._memobj.pmemndx

        def get_prog(i, val_list, valndx, sname, longname, f_apply):
            k = str(i + 1)
            val = val_list[valndx]
            valuelist = RadioSettingValueList(val_list, val)
            rs = RadioSetting(sname + k, longname + k, valuelist)
            rs.set_apply_callback(f_apply, i, self._memobj)
            group.append(rs)
        for i in range(0, self.Pkeys):
            get_prog(i, self.USAGE_LIST,  _progctl[i].usage,
                     "P", "Programmable key ",  self.apply_P)
            get_prog(i, self.SETMODES, _progctl[i].parm, "modeP",
                     "mode for Programmable key ",  self.apply_Pmode)
            get_prog(i, ["0", "1", "2", "3"], _progctl[i].submode, "submodeP",
                     "submode for Programmable key ",  self.apply_Psubmode)
            get_prog(i, self.MEMLIST, _progndx[i], "memP",
                     "mem for Programmable key ",  self.apply_Pmem)
    # ------------ End of special settings handlers.

    # list of group description tuples: [groupame,group title, [param list]].
    # A param is a tuple:
    #  for a simple param: (paramname, paramtitle,[valuename list])
    #  for a handler param: (paramname, paramtitle,( handler,[handler params]))
    # This is a class variable. subclasses must create a variable named
    # class_group_descs. The FT-4 classes simply equate this, but the
    # FT-65 classes must copy and modify this.
    group_descriptions = [
        ("misc", "Miscellaneous Settings", [    # misc
         ("apo", "Automatic Power Off",
          ["OFF"] + ["%0.1f" % (x * 0.5) for x in range(1, 24 + 1)]),
         ("bclo", "Busy Channel Lock-Out", ["OFF", "ON"]),
         ("beep", "Enable the Beeper", ["KEY+SC", "KEY", "OFF"]),
         ("bsy_led", "Busy LED", ["ON", "OFF"]),
         ("edg_beep", "Band Edge Beeper", ["OFF", "ON"]),
         ("vox", "VOX", ["OFF", "ON"]),
         ("rf_squelch", "RF Squelch Threshold",
          ["OFF", "S-1", "S-2", "S-3", "S-4", "S-5", "S-6", "S-7", "S-FULL"]),
         ("tot", "Timeout Timer",
          ["OFF"] + ["%dMIN" % (x) for x in range(1, 30 + 1)]),
         ("tx_led", "TX LED", ["OFF", "ON"]),
         ("use_cwid", "use CW ID", ["NO", "YES"]),
         ("cw_id",  "CW ID Callsign",  (get_strset, [CW_ID_CHARS])),  # handler
         ("vfo_spl", "VFO Split", ["OFF", "ON"]),
         ("wfm_rcv", "Enable Broadband FM", ["OFF", "ON"]),
         ("passwd", "Password",  (get_strset, [PASSWD_CHARS]))  # handler
         ]),
        ("arts", "ARTS Settings", [  # arts
         ("arts_beep", "ARTS BEEP", ["OFF", "INRANG", "ALWAYS"]),
         ("arts_intv", "ARTS Polling Interval", ["25 SEC", "15 SEC"])
         ]),
        ("ctcss", "CTCSS/DCS/DTMF Settings", [  # ctcss
         ("bell", "Bell Repetitions", ["OFF", "1T", "3T", "5T", "8T", "CONT"]),
         ("dtmf_mode", "DTMF Mode", ["Manual", "Auto"]),
         ("dtmf_delay", "DTMF Autodialer Delay Time",
          ["50 ms", "100 ms", "250 ms", "450 ms", "750 ms", "1000 ms"]),
         ("dtmf_speed", "DTMF Autodialer Sending Speed", ["50 ms", "100 ms"]),
         ("dtmf", "DTMF Autodialer Memory ",  (get_dtmfs, []))  # handler
         ]),
        ("switch", "Switch/Knob Settings", [  # switch
         ("lamp", "Lamp Mode", ["5SEC", "10SEC", "30SEC", "KEY", "OFF"]),
         ("moni_tcall", "MONI Switch Function",
          ["MONI", "1750", "2100", "1000", "1450"]),
         ("key_lock", "Lock Function", ["KEY", "PTT", "KEY+PTT"]),
         ("Pkeys", "Pkey fields", (get_progs, []))
         ]),
        ("scan", "Scan Settings", [   # scan
         ("scan_resume", "Scan Resume Mode", ["BUSY", "HOLD", "TIME"]),
         ("pri_rvt", "Priority Revert", ["OFF", "ON"]),
         ("scan_lamp", "Scan Lamp", ["OFF", "ON"]),
         ("wx_alert", "Weather Alert Scan", ["OFF", "ON"])
         ]),
        ("power", "Power Saver Settings", [  # power
         ("battsave", "Receive Mode Battery Save Interval",
          ["OFF", "200 ms", "300 ms", "500 ms", "1 s", "2 s"]),
         ("tx_save", "Transmitter Battery Saver", ["OFF", "ON"])
         ]),
        ("eai", "EAI/EPCS Settings", [  # eai
         ("pager_tx1", "TX pager frequency 1", EPCS_CODES),
         ("pager_tx2", "TX pager frequency 2", EPCS_CODES),
         ("pager_rx1", "RX pager frequency 1", EPCS_CODES),
         ("pager_rx2", "RX pager frequency 2", EPCS_CODES),
         ("pager_ack", "Pager answerback", ["NO", "YES"])
         ])
        ]
    # ----------------end of group_descriptions

    # returns the current values of all the settings in the radio memory image,
    # in the form of a RadioSettings list. Uses the class_group_descs
    # list to create the groups and params. Valuelist scalars are handled
    # inline. More complex params are built by calling the special handlers.
    def get_settings(self):
        _settings = self._memobj.settings
        groups = RadioSettings()
        for description in self.class_group_descs:
            groupname, title, parms = description
            group = RadioSettingGroup(groupname, title)
            groups.append(group)
            for parm in parms:
                param, title, opts = parm
                try:
                    if isinstance(opts, list):
                        # setting is a single value from the list
                        objval = getattr(_settings, param)
                        value = opts[objval]
                        valuelist = RadioSettingValueList(opts, value)
                        group.append(RadioSetting(param, title, valuelist))
                    else:
                        # setting  needs special handling. opts[0] is a
                        # function name
                        opts[0](self,  group, parm)
                except Exception as e:
                    LOG.debug(
                        "%s: cannot set %s to %s" % (e, param, repr(objval))
                        )
        return groups
        # end of get_settings

    # modify settings values in the radio memory image
    def set_settings(self, uisettings):
        _settings = self._memobj.settings
        for element in uisettings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            if not element.changed():
                continue

            try:
                name = element.get_name()
                value = element.value

                if element.has_apply_callback():
                    LOG.debug("Using apply callback")
                    element.run_apply_callback()
                else:
                    setattr(_settings, name, value)

                LOG.debug("Setting %s: %s" % (name, value))
            except Exception:
                LOG.debug(element.get_name())
                raise

    # maps a boolean pair (tx==0,rx==0) to the numbers 0-3
    LOOKUP = [[True, True], [True, False], [False, True], [False, False]]

    def decode_sql(self, mem, chan):
        """
        examine the radio channel fields and determine the correct
        CHIRP CSV values for tmode, cross_mode, and sql_override
        """
        mem.extra = RadioSettingGroup("Extra", "extra")
        extra_modes = ["(None)", "PAGER"]
        value = extra_modes[chan.sql_type == 6]
        valuelist = RadioSettingValueList(extra_modes, value)
        rs = RadioSetting("sql_override", "Squelch override", valuelist)
        mem.extra.append(rs)
        if chan.sql_type == 6:
            return
        sql_map = RADIO_TMODES[chan.sql_type]
        ndx = 0
        if len(sql_map[0]) > 1:
            # the sql_type is TSQL or DCS, so there are multiple UI mappings
            x = getattr(chan, sql_map[1])
            r = getattr(chan, sql_map[2])
            ndx = self.LOOKUP.index([x == 0, r == 0])
            if ndx == 3 and x == r:
                ndx = 4
        mem.tmode = sql_map[0][ndx][0]
        cross = sql_map[0][ndx][1]
        if cross:
            mem.cross_mode = cross
        if chan.rx_ctcss:
            mem.ctone = TONE_MAP[chan.rx_ctcss]
        if chan.tx_ctcss:
            mem.rtone = TONE_MAP[chan.tx_ctcss]
        if chan.tx_dcs:
            mem.dtcs = DTCS_MAP[chan.tx_dcs]
        if chan.rx_dcs:
            mem.rx_dtcs = DTCS_MAP[chan.rx_dcs]

    def encode_sql(self, mem, chan):
        """
        examine CHIRP's mem.tmode and mem.cross_mode and set the values
        for the radio sql_type, dcs codes, and ctcss codes. We set all four
        codes, and then zero out a code if needed when Tone or DCS is one-way
        """
        chan.tx_ctcss = TONE_MAP.index(mem.rtone)
        chan.tx_dcs = DTCS_MAP.index(mem.dtcs)
        chan.rx_ctcss = TONE_MAP.index(mem.ctone)
        chan.rx_dcs = DTCS_MAP.index(mem.rx_dtcs)
        if mem.tmode == "TSQL":
            chan.tx_ctcss = chan.rx_ctcss  # CHIRP uses ctone for TSQL
        if mem.tmode == "DTCS":
            chan.tx_dcs = chan.rx_dcs     # CHIRP uses rx_dtcs for DTCS
        # select the correct internal dictionary and key
        mode_dict, key = [
            (TONE_DICT, mem.tmode),
            (CROSS_DICT, mem.cross_mode)
            ][mem.tmode == "Cross"]
        # now look up that key in that dictionary.
        chan.sql_type, suppress = mode_dict[key]
        if suppress:
            setattr(chan, suppress, 0)
        for setting in mem.extra:
            if (setting.get_name() == 'sql_override'):
                value = str(setting.value)
                if value == "PAGER":
                    chan.sql_type = 6
        return

    # given a CHIRP memory ref, get the radio memobj for it.
    # A memref is either a number or the name of a special
    # CHIRP will sometimes use numbers (>MAX_SLOTS) for specials
    # returns the obj and several attributes
    def slotloc(self, memref):
        array = None
        num = memref
        sname = memref
        if isinstance(memref, str):   # named special?
            num = self.MAX_MEM_SLOT + 1
            for x in self.class_specials:
                try:
                    ndx = x[1].index(memref)
                    array = x[0]
                    break
                except Exception:
                    num += len(x[1])
            if array is None:
                LOG.debug("unknown Special %s", memref)
                raise
            num += ndx
        elif memref > self.MAX_MEM_SLOT:    # numbered special?
            ndx = memref - (self.MAX_MEM_SLOT + 1)
            for x in self.class_specials:
                if ndx < len(x[1]):
                    array = x[0]
                    sname = x[1][ndx]
                    break
                ndx -= len(x[1])
            if array is None:
                LOG.debug("memref number %d out of range", memref)
                raise
        else:                    # regular memory slot
            array = "memory"
            ndx = memref - 1
        memloc = getattr(self._memobj, array)[ndx]
        return (memloc, ndx, num, array, sname)
        # end of slotloc

    # return the raw info for a memory channel
    def get_raw_memory(self, memref):
        memloc, ndx, num, regtype, sname = self.slotloc(memref)
        if regtype == "memory":
            return repr(memloc)
        else:
            return repr(memloc) + repr(self._memobj.names[ndx])

    # return the info for a memory channel In CHIRP canonical form
    def get_memory(self, memref):

        def clean_name(obj):     # helper func to tidy up the name
            name = ''
            for x in range(0, self.namelen):
                y = obj[x]
                if y == 0:
                    break
                if y == 0x7F:    # when programmed from VFO
                    y = 0x20
                name += chr(y)
            return name.rstrip()

        def get_duplex(freq):    # auto duplex to real duplex
            """
            Select the duplex direction if duplex == 'auto'.
            0 is +, 2 is -, and 4 is none.
            """
            return_value = 4     # off, if not in auto range
            for x in self.DUPLEX_AUTO:
                if freq in range(x[0], x[1]):
                    return_value = x[2]
            return return_value

        mem = chirp_common.Memory()
        _mem, ndx, num, regtype, sname = self.slotloc(memref)
        mem.number = num
        freq_offset_factor = self.freq_offset_factor
        # FT-25R/65R Asia version (US is 0)?
        if self.subtype == 3 and freq_offset_factor == 2:
            freq_offset_factor = 1  # 25000 scaler

        # First, we need to know whether a channel is enabled,
        # then we can process any channel parameters.
        # It was found (at least on an FT-25) that channels might be
        # uninitialized and memory is just completely filled with 0xFF.

        if regtype == "pms":
            mem.extd_number = sname
        if regtype in ["memory", "pms"]:
            ndx = num - 1
            mem.name = clean_name(self._memobj.names[ndx].chrs)
            mem.empty = not retrieve_bit(self._memobj.enable, ndx)
            mem.skip = SKIPS[retrieve_bit(self._memobj.scan, ndx)]
        else:
            mem.empty = False
            mem.extd_number = sname
            mem.immutable = ["number", "extd_number", "name", "skip"]

        # So, now if channel is not empty, we can do the evaluation of
        # all parameters. Otherwise we set them to defaults.

        if mem.empty:
            mem.freq = 0
            mem.offset = 0
            mem.duplex = "off"
            mem.power = POWER_LEVELS[0]  # "High"
            mem.mode = "FM"
            mem.tuning_step = 0
        else:
            mem.freq = int(_mem.freq) * 10
            txfreq = int(self._memobj.txfreqs[ndx].freq) * 10
            if _mem.duplex == 5:  # auto offset
                mem.duplex = DUPLEX[get_duplex(mem.freq)]
            else:
                mem.duplex = DUPLEX[_mem.duplex]
            if _mem.duplex == 6:  # split: offset is tx frequency
                mem.offset = txfreq
            else:
                mem.offset = int(_mem.offset) * 25000 * freq_offset_factor
            self.decode_sql(mem, _mem)
            mem.power = POWER_LEVELS[2 - _mem.tx_pwr]
            mem.mode = ["FM", "NFM"][_mem.tx_width]
            mem.tuning_step = STEP_CODE[_mem.step]
        return mem

    def enforce_band(self, memloc, freq, mem_num, sname):
        """
        vfo and home channels are each restricted to a particular band.
        If the frequency is not in the band, use the lower bound
        Raise an exception to cause UI to pop up an error message
        """
        first_vfo_num = self.MAX_MEM_SLOT + len(PMSNAMES) + 1
        band = self.BAND_ASSIGNMENTS[mem_num - first_vfo_num]
        frange = self.valid_bands[band]
        if freq >= frange[0] and freq <= frange[1]:
            memloc.freq = freq / 10
            return freq
        memloc.freq = frange[0] / 10
        raise Exception("freq out of range for %s" % sname)

    # modify a radio channel in memobj based on info in CHIRP canonical form
    def set_memory(self, mem):
        _mem, ndx, num, regtype, sname = self.slotloc(mem.number)
        assert _mem
        if mem.empty:
            if regtype in ["memory", "pms"]:
                store_bit(self._memobj.enable, ndx, False)
                return

        txfreq = mem.freq / 10     # really. RX freq is used for TX base
        _mem.freq = txfreq
        self.encode_sql(mem, _mem)
        if mem.power:
            _mem.tx_pwr = 2 - POWER_LEVELS.index(mem.power)
        else:
            _mem.tx_pwr = 0  # set to "High" if CHIRP canonical value is None
        _mem.tx_width = mem.mode == "NFM"
        _mem.step = STEP_CODE.index(mem.tuning_step)

        freq_offset_factor = self.freq_offset_factor
        # FT-25R/65R Asia version (US version is 0)?
        if self.subtype == 3 and freq_offset_factor == 2:
            freq_offset_factor = 1  # 25000 scaler
        _mem.offset = mem.offset / (25000 * freq_offset_factor)
        duplex = mem.duplex
        if regtype in ["memory", "pms"]:
            ndx = num - 1
            store_bit(self._memobj.enable, ndx, True)
            store_bit(self._memobj.scan, ndx, SKIPS.index(mem.skip))
            nametrim = (mem.name + "        ")[:8]
            self._memobj.names[ndx].chrs = bytearray(nametrim, "ascii")
            if mem.duplex == "split":
                txfreq = mem.offset / 10
            self._memobj.txfreqs[num-1].freq = txfreq
        _mem.duplex = DUPLEX.index(duplex)
        if regtype in ["vfo", "home"]:
            self.enforce_band(_mem, mem.freq, num, sname)

        return


class YaesuFT4GenericRadio(YaesuSC35GenericRadio):
    """
    FT-4 sub family class. Classes for individual radios extend
    these classes and are found at the end of this file.
    """
    class_specials = SPECIALS_FT4
    Pkeys = 2     # number of programmable keys
    namelen = 6   # length of the mem name display on the front-panel
    freq_offset_factor = 1  # 25000 * 1
    class_group_descs = YaesuSC35GenericRadio.group_descriptions
    # names for the setmode function for the programmable keys. Mode zero means
    # that the key is programmed for a memory not a setmode.
    SETMODES = [
        "mem", "apo", "ar bep", "ar int", "beclo",              # 00-04
        "beep", "bell", "cw id", "cw wrt", "dc vlt",            # 05-09
        "dcs cod", "dt dly", "dt set", "dtc spd", "edg.bep",    # 10-14
        "lamp", "led.bsy", "led.tx", "lock", "m/t-cl",          # 15-19
        "mem.del", "mem.tag", "pag.abk", "pag.cdr", "pag.cdt",  # 20-24
        "pri.rvt", "pswd", "pswdwt", "rf sql", "rpt.ars",       # 25-29
        "rpt.frq", "rpt.sft", "rxsave", "scn.lmp", "scn.rsm",   # 30-34
        "skip", "sql.typ", "step", "tn frq", "tot",             # 35-39
        "tx pwr", "tx save", "vfo.spl", "vox", "wfm.rcv",       # 40-44
        "w/n.dev", "wx.alert"                                   # 45-46
        ]


class YaesuFT65GenericRadio(YaesuSC35GenericRadio):
    """
    FT-65 sub family class. Classes for individual radios extend
    these classes and are found at the end of this file.
    """
    class_specials = SPECIALS_FT65
    Pkeys = 4     # number of programmable keys
    namelen = 8   # length of the mem name display on the front-panel
    freq_offset_factor = 2  # 25000 * 2
    # we need a deep copy here because we are adding deeper than the top level.
    class_group_descs = copy.deepcopy(YaesuSC35GenericRadio.group_descriptions)
    add_paramdesc(
        class_group_descs, "misc", ("compander", "Compander", ["OFF", "ON"]))
    # names for the setmode function for the programmable keys. Mode zero means
    # that the key is programmed for a memory not a setmode.
    SETMODES = [
        "mem", "apo", "arts", "battsave", "b-ch.l/o",              # 00-04
        "beep", "bell", "compander", "ctcss", "cw id",             # 05-09
        "dc volt", "dcs code", "dtmf set", "dtmf wrt", "edg bep",  # 10-14
        "key lock", "lamp", "ledbsy", "mem del", "mon/t-cl",       # 15-19
        "name tag", "pager", "password", "pri.rvt", "repeater",    # 20-24
        "resume", "rf.sql", "scn.lamp", "skip", "sql type",        # 25-29
        "step", "tot", "tx pwr", "tx save", "vfo.spl",             # 30-34
        "vox", "wfm.rcv", "wide/nar", "wx alert", "scramble"       # 35-39
        ]


# Classes for each individual radio.


@directory.register
class YaesuFT4XRRadio(YaesuFT4GenericRadio):
    """
    FT-4X dual band, US version
    """
    MODEL = "FT-4XR"
    id_str = b'IFT-35R\x00\x00V100\x00\x00'
    valid_bands = VALID_BANDS_DUAL
    DUPLEX_AUTO = DUPLEX_AUTO_US
    legal_steps = US_LEGAL_STEPS
    BAND_ASSIGNMENTS = BAND_ASSIGNMENTS_DUALBAND


@directory.register
class YaesuFT4XERadio(YaesuFT4GenericRadio):
    """
    FT-4X dual band, EU version
    """
    MODEL = "FT-4XE"
    id_str = b'IFT-35R\x00\x00V100\x00\x00'
    valid_bands = VALID_BANDS_DUAL
    DUPLEX_AUTO = DUPLEX_AUTO_EU
    legal_steps = STEP_CODE
    BAND_ASSIGNMENTS = BAND_ASSIGNMENTS_DUALBAND


@directory.register
class YaesuFT4VRRadio(YaesuFT4GenericRadio):
    """
    FT-4V VHF, US version
    """
    MODEL = "FT-4VR"
    id_str = b'IFT-15R\x00\x00V100\x00\x00'
    valid_bands = VALID_BANDS_VHF
    DUPLEX_AUTO = DUPLEX_AUTO_US
    legal_steps = US_LEGAL_STEPS
    BAND_ASSIGNMENTS = BAND_ASSIGNMENTS_MONO_VHF


# No image available yet
# @directory.register
# class YaesuFT4VERadio(YaesuFT4GenericRadio):
#    """
#    FT-4V VHF, EU version
#    """
#    MODEL = "FT-4VE"
#    id_str = b'IFT-15R\x00\x00V100\x00\x00'
#    valid_bands = VALID_BANDS_VHF
#    DUPLEX_AUTO = DUPLEX_AUTO_EU
#    legal_steps = STEP_CODE
#    BAND_ASSIGNMENTS = BAND_ASSIGNMENTS_MONO_VHF


@directory.register
class YaesuFT65RRadio(YaesuFT65GenericRadio):
    """
    FT-65 dual band, US/Asia version
    """
    MODEL = "FT-65R"
    id_str = b'IH-420\x00\x00\x00V100\x00\x00'
    valid_bands = VALID_BANDS_DUAL
    DUPLEX_AUTO = DUPLEX_AUTO_US
    legal_steps = STEP_CODE
    BAND_ASSIGNMENTS = BAND_ASSIGNMENTS_DUALBAND


@directory.register
class YaesuFT65ERadio(YaesuFT65GenericRadio):
    """
    FT-65 dual band, EU version
    """
    MODEL = "FT-65E"
    id_str = b'IH-420\x00\x00\x00V100\x00\x00'
    valid_bands = VALID_BANDS_DUAL
    DUPLEX_AUTO = DUPLEX_AUTO_EU
    legal_steps = STEP_CODE
    BAND_ASSIGNMENTS = BAND_ASSIGNMENTS_DUALBAND


@directory.register
class YaesuFT25RRadio(YaesuFT65GenericRadio):
    """
    FT-25 VHF, US/Asia version
    """
    MODEL = "FT-25R"
    id_str = b'IFT-25R\x00\x00V100\x00\x00'
    valid_bands = VALID_BANDS_VHF
    DUPLEX_AUTO = DUPLEX_AUTO_US
    legal_steps = US_LEGAL_STEPS
    BAND_ASSIGNMENTS = BAND_ASSIGNMENTS_MONO_VHF


# No image available yet
# @directory.register
# class YaesuFT25ERadio(YaesuFT65GenericRadio):
#    """
#    FT-25 VHF, EU version
#    """
#    MODEL = "FT-25E"
#    id_str = b'IFT-25R\x00\x00V100\x00\x00'
#    valid_bands = VALID_BANDS_VHF
#    DUPLEX_AUTO = DUPLEX_AUTO_EU
#    legal_steps = STEP_CODE
#    BAND_ASSIGNMENTS = BAND_ASSIGNMENTS_MONO_VHF



# Copyright 2017 Windsor Schmidt <windsor.schmidt@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp import chirp_common, directory, bitwise
from chirp.drivers import icf
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueList, RadioSettingValueBoolean, RadioSettings

# The Icom IC-2300H is a 65W, 144 MHz mobile transceiver based on the IC-2200H.
# Unlike the IC-2200H, this model does not accept Icom's UT-118 D-STAR board.
#
# A simple USB interface based on a typical FT232RL breakout board was used
# during development of this module. A schematic diagram is as follows:
#
#
# 3.5mm plug from IC-2300H
# sleeve / ring / tip
# --.______________
#   |      |   |   \
#   |______|___|___/    FT232RL breakout
# --'    |   |        .------------------.
#        |   +--------| RXD              |
#        |   |   D1   |                  |
#        |   +--|>|---| TXD              | USB/PC
#        |   |   R1   |                  |-------->
#        |   +--[_]---| VCC (5V)         |
#        |            |                  |
#        +------------| GND              |
#                     `------------------'
#
# D1: 1N4148 Schottky diode
# R1: 10K ohm resistor

MEM_FORMAT = """
#seekto 0x0000; // channel memories
struct {
  ul16 frequency;
  ul16 offset;
  char name[6];
  u8   repeater_tone;
  u8   ctcss_tone;
  u8   dtcs_code;
  u8   tuning_step:4,
       tone_mode:4;
  u8   unknown1:3,
       mode_narrow:1,
       power:2,
       unknown2:2;
  u8   dtcs_polarity:2,
       duplex:2,
       unknown3:1,
       reverse_duplex:1,
       tx_inhibit:1,
       display_style:1;
} memory[200];
#seekto 0x1340; // channel memory flags
struct {
  u8   unknown5:2,
       empty:1,
       skip:1,
       bank:4;
} flags[200];
#seekto 0x1660; // power-on and regular set menu items
struct {
  u8   key_beep;
  u8   tx_timeout;
  u8   auto_repeater;
  u8   auto_power_off;
  u8   repeater_lockout;
  u8   squelch_delay;
  u8   squelch_type;
  u8   dtmf_speed;
  u8   display_type;
  u8   unknown6;
  u8   tone_burst;
  u8   voltage_display;
  u8   unknown7;
  u8   display_brightness;
  u8   display_color;
  u8   auto_dimmer;
  u8   display_contrast;
  u8   scan_pause_timer;
  u8   mic_gain;
  u8   scan_resume_timer;
  u8   weather_alert;
  u8   bank_link_enable;
  u8   bank_link[10];
} settings;
"""

TUNING_STEPS = [5.0, 6.25, 10.0, 12.5, 15.0, 20.0, 25.0, 30.0, 50.0]
TONE_MODES = ["", "Tone", "TSQL", "DTCS"]
DUPLEX = ["", "-", "+"]
DTCSP = ["NN", "NR", "RN", "RR"]
DTCS_POLARITY = ["NN", "NR", "RN", "RR"]

POWER_LEVELS = [chirp_common.PowerLevel("High", watts=65),
                chirp_common.PowerLevel("Low", watts=5),
                chirp_common.PowerLevel("MidLow", watts=10),
                chirp_common.PowerLevel("Mid", watts=25)]


def _wipe_memory(mem, char):
    mem.set_raw(char * (mem.size() // 8))


@directory.register
class IC2300Radio(icf.IcomCloneModeRadio):
    """Icom IC-2300"""
    VENDOR = "Icom"
    MODEL = "IC-2300H"

    _model = "\x32\x51\x00\x01"
    _memsize = 6304
    _endframe = "Icom Inc.C5\xfd"
    _can_hispeed = True
    _ranges = [(0x0000, 0x18a0, 32)]  # upload entire memory for now

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.memory_bounds = (0, 199)
        rf.valid_modes = ["FM", "NFM"]
        rf.valid_tmodes = list(TONE_MODES)
        rf.valid_duplexes = list(DUPLEX)
        rf.valid_tuning_steps = list(TUNING_STEPS)
        rf.valid_bands = [(136000000, 174000000)]  # USA tx range: 144-148 MHz
        rf.valid_skips = ["", "S"]
        rf.valid_power_levels = POWER_LEVELS
        rf.has_settings = True
        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def _get_bank(self, loc):
        _flag = self._memobj.flags[loc]
        if _flag.bank == 0x0a:
            return None
        else:
            return _flag.bank

    def _set_bank(self, loc, bank):
        _flag = self._memobj.flags[loc]
        if bank is None:
            _flag.bank = 0x0a
        else:
            _flag.bank = bank

    def get_memory(self, number):
        _mem = self._memobj.memory[number]
        _flag = self._memobj.flags[number]
        mem = chirp_common.Memory()
        mem.number = number
        if _flag.empty:
            mem.empty = True
            return mem
        mult = int(TUNING_STEPS[_mem.tuning_step] * 1000)
        mem.freq = (_mem.frequency * mult)
        mem.offset = (_mem.offset * mult)
        mem.name = str(_mem.name).rstrip()
        mem.rtone = chirp_common.TONES[_mem.repeater_tone]
        mem.ctone = chirp_common.TONES[_mem.ctcss_tone]
        mem.dtcs = chirp_common.DTCS_CODES[_mem.dtcs_code]
        mem.tuning_step = TUNING_STEPS[_mem.tuning_step]
        mem.tmode = TONE_MODES[_mem.tone_mode]
        mem.mode = "NFM" if _mem.mode_narrow else "FM"
        mem.dtcs_polarity = DTCS_POLARITY[_mem.dtcs_polarity]
        mem.duplex = DUPLEX[_mem.duplex]
        mem.skip = "S" if _flag.skip else ""
        mem.power = POWER_LEVELS[_mem.power]

        # Reverse duplex
        mem.extra = RadioSettingGroup("extra", "Extra")
        rev = RadioSetting("reverse_duplex", "Reverse duplex",
                           RadioSettingValueBoolean(bool(_mem.reverse_duplex)))
        rev.set_doc("Reverse duplex")
        mem.extra.append(rev)

        # Tx inhibit
        tx_inhibit = RadioSetting("tx_inhibit", "TX inhibit",
                                  RadioSettingValueBoolean(
                                      bool(_mem.tx_inhibit)))
        tx_inhibit.set_doc("TX inhibit")
        mem.extra.append(tx_inhibit)

        # Memory display style
        opt = ["Frequency", "Label"]
        dsp = RadioSetting("display_style", "Display style",
                           RadioSettingValueList(opt, opt[_mem.display_style]))
        dsp.set_doc("Memory display style")
        mem.extra.append(dsp)

        return mem

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])

    def set_memory(self, mem):
        number = mem.number
        _mem = self._memobj.memory[number]
        _flag = self._memobj.flags[number]
        was_empty = int(_flag.empty)
        _flag.empty = mem.empty
        if mem.empty:
            _wipe_memory(_mem, "\xff")
            return
        if was_empty:
            _wipe_memory(_mem, "\x00")
        mult = mem.tuning_step * 1000
        _mem.frequency = (mem.freq / mult)
        _mem.offset = mem.offset / mult
        _mem.name = mem.name.ljust(6)
        _mem.repeater_tone = chirp_common.TONES.index(mem.rtone)
        _mem.ctcss_tone = chirp_common.TONES.index(mem.ctone)
        _mem.dtcs_code = chirp_common.DTCS_CODES.index(mem.dtcs)
        _mem.tuning_step = TUNING_STEPS.index(mem.tuning_step)
        _mem.tone_mode = TONE_MODES.index(mem.tmode)
        _mem.mode_narrow = mem.mode.startswith("N")
        _mem.dtcs_polarity = DTCSP.index(mem.dtcs_polarity)
        _mem.duplex = DUPLEX.index(mem.duplex)
        if mem.power:
            _mem.power = POWER_LEVELS.index(mem.power)
        else:
            _mem.power = POWER_LEVELS[0]
        _flag.skip = mem.skip != ""

        for setting in mem.extra:
            setattr(_mem, setting.get_name(), setting.value)

    def get_settings(self):
        _settings = self._memobj.settings
        basic = RadioSettingGroup("basic", "Basic Settings")
        front_panel = RadioSettingGroup("front_panel", "Front Panel Settings")
        top = RadioSettings(basic, front_panel)

        # Transmit timeout
        opt = ['Disabled', '1 minute'] + \
              [s + ' minutes' for s in map(str, range(2, 31))]
        rs = RadioSetting("tx_timeout", "Transmit timeout (min)",
                          RadioSettingValueList(opt, opt[
                              _settings.tx_timeout
                          ]))
        basic.append(rs)

        # Auto Repeater (USA model only)
        opt = ["Disabled", "Duplex Only", "Duplex and tone"]
        rs = RadioSetting("auto_repeater", "Auto repeater",
                          RadioSettingValueList(opt, opt[
                              _settings.auto_repeater
                          ]))
        basic.append(rs)

        # Auto Power Off
        opt = ["Disabled", "30 minutes", "60 minutes", "120 minutes"]
        rs = RadioSetting("auto_power_off", "Auto power off",
                          RadioSettingValueList(opt, opt[
                              _settings.auto_power_off
                          ]))
        basic.append(rs)

        # Squelch Delay
        opt = ["Short", "Long"]
        rs = RadioSetting("squelch_delay", "Squelch delay",
                          RadioSettingValueList(opt, opt[
                              _settings.squelch_delay
                          ]))
        basic.append(rs)

        # Squelch Type
        opt = ["Noise squelch", "S-meter squelch", "Squelch attenuator"]
        rs = RadioSetting("squelch_type", "Squelch type",
                          RadioSettingValueList(opt, opt[
                              _settings.squelch_type
                          ]))
        basic.append(rs)

        # Repeater Lockout
        opt = ["Disabled", "Repeater lockout", "Busy lockout"]
        rs = RadioSetting("repeater_lockout", "Repeater lockout",
                          RadioSettingValueList(opt, opt[
                              _settings.repeater_lockout
                          ]))
        basic.append(rs)

        # DTMF Speed
        opt = ["100ms interval, 5.0 cps",
               "200ms interval, 2.5 cps",
               "300ms interval, 1.6 cps",
               "500ms interval, 1.0 cps"]
        rs = RadioSetting("dtmf_speed", "DTMF speed",
                          RadioSettingValueList(opt, opt[
                              _settings.dtmf_speed
                          ]))
        basic.append(rs)

        # Scan pause timer
        opt = [s + ' seconds' for s in map(str, range(2, 22, 2))] + ['Hold']
        rs = RadioSetting("scan_pause_timer", "Scan pause timer",
                          RadioSettingValueList(
                              opt, opt[_settings.scan_pause_timer]))
        basic.append(rs)

        # Scan Resume Timer
        opt = ['Immediate'] + \
              [s + ' seconds' for s in map(str, range(1, 6))] + ['Hold']
        rs = RadioSetting("scan_resume_timer", "Scan resume timer",
                          RadioSettingValueList(
                              opt, opt[_settings.scan_resume_timer]))
        basic.append(rs)

        # Weather Alert (USA model only)
        rs = RadioSetting("weather_alert", "Weather alert",
                          RadioSettingValueBoolean(_settings.weather_alert))
        basic.append(rs)

        # Tone Burst
        rs = RadioSetting("tone_burst", "Tone burst",
                          RadioSettingValueBoolean(_settings.tone_burst))
        basic.append(rs)

        # Memory Display Type
        opt = ["Frequency", "Channel", "Name"]
        rs = RadioSetting("display_type", "Memory display",
                          RadioSettingValueList(opt,
                                                opt[_settings.display_type]))
        front_panel.append(rs)

        # Display backlight brightness;
        opt = ["1 (dimmest)", "2", "3", "4 (brightest)"]
        rs = RadioSetting("display_brightness", "Backlight brightness",
                          RadioSettingValueList(
                              opt,
                              opt[_settings.display_brightness]))
        front_panel.append(rs)

        # Display backlight color
        opt = ["Amber", "Yellow", "Green"]
        rs = RadioSetting("display_color", "Backlight color",
                          RadioSettingValueList(opt,
                                                opt[_settings.display_color]))
        front_panel.append(rs)

        # Display contrast
        opt = ["1 (lightest)", "2", "3", "4 (darkest)"]
        rs = RadioSetting("display_contrast", "Display contrast",
                          RadioSettingValueList(
                              opt,
                              opt[_settings.display_contrast]))
        front_panel.append(rs)

        # Auto dimmer
        opt = ["Disabled", "Backlight off", "1 (dimmest)", "2", "3"]
        rs = RadioSetting("auto_dimmer", "Auto dimmer",
                          RadioSettingValueList(opt,
                                                opt[_settings.auto_dimmer]))
        front_panel.append(rs)

        # Microphone gain
        opt = ["Low", "High"]
        rs = RadioSetting("mic_gain", "Microphone gain",
                          RadioSettingValueList(opt,
                                                opt[_settings.mic_gain]))
        front_panel.append(rs)

        # Key press beep
        rs = RadioSetting("key_beep", "Key press beep",
                          RadioSettingValueBoolean(_settings.key_beep))
        front_panel.append(rs)

        # Voltage Display;
        rs = RadioSetting("voltage_display", "Voltage display",
                          RadioSettingValueBoolean(_settings.voltage_display))
        front_panel.append(rs)

        # TODO: Add Bank Links settings to GUI

        return top

    def set_settings(self, settings):
        _settings = self._memobj.settings
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            setting = element.get_name()
            setattr(_settings, setting, element.value)



# Copyright 2011 Dan Smith <dsmith@danplanet.com>
#
# FT-2900-specific modifications by Richard Cochran, <ag6qr@sonic.net>
# Initial work on settings by Chris Fosnight, <chris.fosnight@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
import logging

from chirp import util, memmap, chirp_common, bitwise, directory, errors
from chirp.drivers.yaesu_clone import YaesuCloneModeRadio
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueList, RadioSettingValueString, RadioSettings


LOG = logging.getLogger(__name__)


def _send(s, data):
    s.write(data)
    echo = s.read(len(data))
    if data != echo:
        raise Exception("Failed to read echo")
    LOG.debug("got echo\n%s\n" % util.hexprint(echo))


ACK = b"\x06"
INITIAL_CHECKSUM = 0


def _download(radio):

    blankChunk = b""
    for _i in range(0, 32):
        blankChunk += b"\xff"

    LOG.debug("in _download\n")

    data = b""
    for _i in range(0, 20):
        data = radio.pipe.read(20)
        LOG.debug("Header:\n%s" % util.hexprint(data))
        LOG.debug("len(header) = %s\n" % len(data))

        if data == radio.IDBLOCK:
            break

    if data != radio.IDBLOCK:
        raise Exception("Failed to read header")

    _send(radio.pipe, ACK)

    # initialize data, the big var that holds all memory
    data = b""

    _blockNum = 0

    while len(data) < radio._block_sizes[1]:
        _blockNum += 1
        time.sleep(0.03)
        chunk = radio.pipe.read(32)
        LOG.debug("Block %i " % (_blockNum))
        if chunk == blankChunk:
            LOG.debug("blank chunk\n")
        else:
            LOG.debug("Got: %i:\n%s" % (len(chunk), util.hexprint(chunk)))
        if len(chunk) != 32:
            LOG.debug("len chunk is %i\n" % (len(chunk)))
            raise Exception("Failed to get full data block")
        else:
            data += chunk

        if radio.status_fn:
            status = chirp_common.Status()
            status.max = radio._block_sizes[1]
            status.cur = len(data)
            status.msg = "Cloning from radio"
            radio.status_fn(status)

    LOG.debug("Total: %i" % len(data))

    # radio should send us one final termination byte, containing
    # checksum
    chunk = radio.pipe.read(32)
    if len(chunk) != 1:
        LOG.debug("len(chunk) is %i\n" % len(chunk))
        raise Exception("radio sent extra unknown data")
    LOG.debug("Got: %i:\n%s" % (len(chunk), util.hexprint(chunk)))

    # compute checksum
    cs = INITIAL_CHECKSUM
    for byte in radio.IDBLOCK:
        cs += byte
    for byte in data:
        cs += byte
    LOG.debug("calculated checksum is %x\n" % (cs & 0xff))
    LOG.debug("Radio sent checksum is %x\n" % chunk[0])

    if (cs & 0xff) != chunk[0]:
        raise Exception("Failed checksum on read.")

    # for debugging purposes, dump the channels, in hex.
    for _i in range(0, 200):
        _startData = 1892 + 20 * _i
        chunk = data[_startData:_startData + 20]
        LOG.debug("channel %i:\n%s" % (_i, util.hexprint(chunk)))

    return memmap.MemoryMapBytes(data)


def _upload(radio):
    for _i in range(0, 10):
        data = radio.pipe.read(256)
        if not data:
            break
        LOG.debug("What is this garbage?\n%s" % util.hexprint(data))
        raise Exception("Radio sent unrecognized data")

    _send(radio.pipe, radio.IDBLOCK)
    time.sleep(.2)
    ack = radio.pipe.read(300)
    LOG.debug("Ack was (%i):\n%s" % (len(ack), util.hexprint(ack)))
    if ack != ACK:
        raise Exception("Radio did not ack ID. Check cable, verify"
                        " radio is not locked.\n"
                        " (press & Hold red \"*L\" button to unlock"
                        " radio if needed)")

    block = 0
    cs = INITIAL_CHECKSUM
    for byte in radio.IDBLOCK:
        cs += byte

    while block < (radio.get_memsize() // 32):
        data = radio.get_mmap()[block * 32:(block + 1) * 32]

        LOG.debug("Writing block %i:\n%s" % (block, util.hexprint(data)))

        _send(radio.pipe, data)
        time.sleep(0.03)

        for byte in data:
            cs += byte

        if radio.status_fn:
            status = chirp_common.Status()
            status.max = radio._block_sizes[1]
            status.cur = block * 32
            status.msg = "Cloning to radio"
            radio.status_fn(status)
        block += 1

    _send(radio.pipe, bytes([cs & 0xFF]))


MEM_FORMAT = """
#seekto 0x0080;
struct {
    u8  apo;
    u8  arts_beep;
    u8  bell;
    u8  dimmer;
    u8  cw_id_string[16];
    u8  cw_trng;
    u8  x95;
    u8  x96;
    u8  x97;
    u8  int_cd;
    u8  int_set;
    u8  x9A;
    u8  x9B;
    u8  lock;
    u8  x9D;
    u8  mic_gain;
    u8  open_msg;
    u8  openMsg_Text[6];
    u8  rf_sql;
    u8  unk:6,
        pag_abk:1,
        unk:1;
    u8  pag_cdr_1;
    u8  pag_cdr_2;
    u8  pag_cdt_1;
    u8  pag_cdt_2;
    u8  prog_p1;
    u8  xAD;
    u8  prog_p2;
    u8  xAF;
    u8  prog_p3;
    u8  xB1;
    u8  prog_p4;
    u8  xB3;
    u8  resume;
    u8  tot;
    u8  unk:1,
        cw_id:1,
        unk:1,
        ts_speed:1,
        ars:1,
        unk:2,
        dtmf_mode:1;
    u8  unk:1,
        ts_mut:1,
        wires_auto:1,
        busy_lockout:1,
        edge_beep:1,
        unk:3;
    u8  unk:2,
        s_search:1,
        unk:2,
        cw_trng_units:1,
        unk:2;
    u8  dtmf_speed:1,
        unk:2,
        arts_interval:1,
        unk:1,
        inverted_dcs:1,
        unk:1,
        mw_mode:1;
    u8  unk:2,
        wires_mode:1,
        wx_alert:1,
        unk:1,
        wx_vol_max:1,
        revert:1,
        unk:1;
    u8  vfo_scan;
    u8  scan_mode;
    u8  dtmf_delay;
    u8  beep;
    u8  xBF;
} settings;

#seekto 0x00d0;
    u8  passwd[4];
    u8  mbs;

#seekto 0x00c0;
struct {
  u16 in_use;
} bank_used[8];

#seekto 0x00ef;
  u8 currentTone;

#seekto 0x00f0;
  u8 curChannelMem[20];

#seekto 0x1e0;
struct {
  u8 dtmf_string[16];
} dtmf_strings[10];

#seekto 0x0127;
  u8 curChannelNum;

#seekto 0x012a;
  u8 banksoff1;

#seekto 0x15f;
  u8 checksum1;

#seekto 0x16f;
  u8 curentTone2;

#seekto 0x1aa;
  u16 banksoff2;

#seekto 0x1df;
  u8 checksum2;

#seekto 0x0360;
struct{
  u8 name[6];
} bank_names[8];


#seekto 0x03c4;
struct{
  u16 channels[50];
} banks[8];

#seekto 0x06e4;
struct {
  u8 even_pskip:1,
     even_skip:1,
     even_valid:1,
     even_masked:1,
     odd_pskip:1,
     odd_skip:1,
     odd_valid:1,
     odd_masked:1;
} flags[225];

#seekto 0x0764;
struct {
  u8 unknown0:2,
     isnarrow:1,
     unknown1:5;
  u8 unknown2:2,
     duplex:2,
     unknown3:1,
     step:3;
  bbcd freq[3];
  u8 power:2,
     unknown4:3,
     tmode:3;
  u8 name[6];
  bbcd offset[3];
  u8 ctonesplitflag:1,
     ctone:7;
  u8 rx_dtcssplitflag:1,
     rx_dtcs:7;
  u8 unknown5;
  u8 rtonesplitflag:1,
     rtone:7;
  u8 dtcssplitflag:1,
     dtcs:7;
} memory[200];

"""

MODES = ["FM", "NFM"]
TMODES = ["", "Tone", "TSQL", "DTCS", "TSQL-R", "Cross"]
CROSS_MODES = ["DTCS->", "Tone->DTCS", "DTCS->Tone",
               "Tone->Tone", "DTCS->DTCS"]
DUPLEX = ["", "-", "+", "split"]
POWER_LEVELS = [chirp_common.PowerLevel("Hi", watts=75),
                chirp_common.PowerLevel("Low3", watts=30),
                chirp_common.PowerLevel("Low2", watts=10),
                chirp_common.PowerLevel("Low1", watts=5),
                ]

CHARSET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ +-/?C[] _"
STEPS = [5.0, 10.0, 12.5, 15.0, 20.0, 25.0, 50.0, 100.0]


def _decode_tone(radiotone):
    try:
        chirptone = chirp_common.TONES[radiotone]
    except IndexError:
        chirptone = 100
        LOG.debug("found invalid radio tone: %i\n" % radiotone)
    return chirptone


def _decode_dtcs(radiodtcs):
    try:
        chirpdtcs = chirp_common.DTCS_CODES[radiodtcs]
    except IndexError:
        chirpdtcs = 23
        LOG.debug("found invalid radio dtcs code: %i\n" % radiodtcs)
    return chirpdtcs


def _decode_name(mem):
    name = ""
    for i in mem:
        if (i & 0x7F) == 0x7F:
            break
        try:
            name += CHARSET[i & 0x7F]
        except IndexError:
            LOG.debug("Unknown char index: %x " % (i))
    name = name.strip()
    return name


def _encode_name(mem):
    if (mem.strip() == ""):
        return [0xff] * 6

    name = [None] * 6
    for i in range(0, 6):
        try:
            name[i] = CHARSET.index(mem[i])
        except IndexError:
            name[i] = CHARSET.index(" ")

    name[0] = name[0] | 0x80
    return name


def _wipe_memory(mem):
    mem.set_raw(b"\xff" * (mem.size() // 8))


class FT2900Bank(chirp_common.NamedBank):

    def get_name(self):
        _bank = self._model._radio._memobj.bank_names[self.index]
        name = ""
        for i in _bank.name:
            if i == 0xff:
                break
            name += CHARSET[i & 0x7f]

        return name.rstrip()

    def set_name(self, name):
        name = name.upper().ljust(6)[:6]
        _bank = self._model._radio._memobj.bank_names[self.index]
        _bank.name = [CHARSET.index(x) for x in name.ljust(6)[:6]]


class FT2900BankModel(chirp_common.BankModel):

    def get_num_mappings(self):
        return 8

    def get_mappings(self):
        banks = self._radio._memobj.banks
        bank_mappings = []
        for index, _bank in enumerate(banks):
            bank = FT2900Bank(self, "%i" % index, "b%i" % (index + 1))
            bank.index = index
            bank_mappings.append(bank)

        return bank_mappings

    def _get_channel_numbers_in_bank(self, bank):
        _bank_used = self._radio._memobj.bank_used[bank.index]
        if _bank_used.in_use == 0xffff:
            return set()

        _members = self._radio._memobj.banks[bank.index]
        return set([int(ch) for ch in _members.channels if ch != 0xffff])

    def _update_bank_with_channel_numbers(self, bank, channels_in_bank):
        _members = self._radio._memobj.banks[bank.index]
        if len(channels_in_bank) > len(_members.channels):
            raise Exception("More than %i entries in bank %d" %
                            (len(_members.channels), bank.index))

        empty = 0
        for index, channel_number in enumerate(sorted(channels_in_bank)):
            _members.channels[index] = channel_number
            empty = index + 1
        for index in range(empty, len(_members.channels)):
            _members.channels[index] = 0xffff

        _bank_used = self._radio._memobj.bank_used[bank.index]
        if empty == 0:
            _bank_used.in_use = 0xffff
        else:
            _bank_used.in_use = empty - 1

    def add_memory_to_mapping(self, memory, bank):
        channels_in_bank = self._get_channel_numbers_in_bank(bank)
        channels_in_bank.add(memory.number)
        self._update_bank_with_channel_numbers(bank, channels_in_bank)

        # tells radio that banks are active
        self._radio._memobj.banksoff1 = bank.index
        self._radio._memobj.banksoff2 = bank.index

    def remove_memory_from_mapping(self, memory, bank):
        channels_in_bank = self._get_channel_numbers_in_bank(bank)
        try:
            channels_in_bank.remove(memory.number)
        except KeyError:
            raise Exception("Memory %i is not in bank %s. Cannot remove" %
                            (memory.number, bank))
        self._update_bank_with_channel_numbers(bank, channels_in_bank)

    def get_mapping_memories(self, bank):
        memories = []
        for channel in self._get_channel_numbers_in_bank(bank):
            memories.append(self._radio.get_memory(channel))

        return memories

    def get_memory_mappings(self, memory):
        banks = []
        for bank in self.get_mappings():
            if memory.number in self._get_channel_numbers_in_bank(bank):
                banks.append(bank)

        return banks


@directory.register
class FT2900Radio(YaesuCloneModeRadio):

    """Yaesu FT-2900"""
    VENDOR = "Yaesu"
    MODEL = "FT-2900R/1900R"
    IDBLOCK = b"\x56\x43\x32\x33\x00\x02\x46\x01\x01\x01"
    BAUD_RATE = 19200

    _memsize = 8000
    _block_sizes = [8, 8000]

    def get_features(self):
        rf = chirp_common.RadioFeatures()

        rf.memory_bounds = (0, 199)

        rf.can_odd_split = True
        rf.has_ctone = True
        rf.has_rx_dtcs = True
        rf.has_cross = True
        rf.has_dtcs_polarity = False
        rf.has_bank = True
        rf.has_bank_names = True
        rf.has_settings = True

        rf.valid_tuning_steps = STEPS
        rf.valid_modes = MODES
        rf.valid_tmodes = TMODES
        rf.valid_cross_modes = CROSS_MODES
        rf.valid_bands = [(136000000, 174000000)]
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_duplexes = DUPLEX
        rf.valid_skips = ["", "S", "P"]
        rf.valid_name_length = 6
        rf.valid_characters = CHARSET

        return rf

    def sync_in(self):
        start = time.time()
        try:
            self._mmap = _download(self)
        except errors.RadioError:
            raise
        except Exception as e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)
        LOG.info("Downloaded in %.2f sec" % (time.time() - start))
        self.process_mmap()

    def sync_out(self):
        self.pipe.timeout = 1
        start = time.time()
        try:
            _upload(self)
        except errors.RadioError:
            raise
        except Exception as e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)
        LOG.info("Uploaded in %.2f sec" % (time.time() - start))

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])

    def get_memory(self, number):
        _mem = self._memobj.memory[number]
        _flag = self._memobj.flags[(number) / 2]

        nibble = ((number) % 2) and "even" or "odd"
        used = _flag["%s_masked" % nibble]
        valid = _flag["%s_valid" % nibble]
        pskip = _flag["%s_pskip" % nibble]
        skip = _flag["%s_skip" % nibble]

        mem = chirp_common.Memory()

        mem.number = number

        if _mem.get_raw(asbytes=False)[0] == "\xFF" or not valid or not used:
            mem.empty = True
            return mem

        mem.tuning_step = STEPS[_mem.step]
        mem.freq = int(_mem.freq) * 1000

        # compensate for 12.5 kHz tuning steps, add 500 Hz if needed
        if (mem.tuning_step == 12.5):
            lastdigit = int(_mem.freq) % 10
            if (lastdigit == 2 or lastdigit == 7):
                mem.freq += 500

        mem.offset = chirp_common.fix_rounded_step(int(_mem.offset) * 1000)
        mem.duplex = DUPLEX[_mem.duplex]
        if _mem.tmode < TMODES.index("Cross"):
            mem.tmode = TMODES[_mem.tmode]
            mem.cross_mode = CROSS_MODES[0]
        else:
            mem.tmode = "Cross"
            mem.cross_mode = CROSS_MODES[_mem.tmode - TMODES.index("Cross")]

        mem.rtone = _decode_tone(_mem.rtone)
        mem.ctone = _decode_tone(_mem.ctone)

        # check for unequal ctone/rtone in TSQL mode.  map it as a
        # cross tone mode
        if mem.rtone != mem.ctone and (mem.tmode == "TSQL" or
                                       mem.tmode == "Tone"):
            mem.tmode = "Cross"
            mem.cross_mode = "Tone->Tone"

        mem.dtcs = _decode_dtcs(_mem.dtcs)
        mem.rx_dtcs = _decode_dtcs(_mem.rx_dtcs)

        # check for unequal dtcs/rx_dtcs in DTCS mode.  map it as a
        # cross tone mode
        if mem.dtcs != mem.rx_dtcs and mem.tmode == "DTCS":
            mem.tmode = "Cross"
            mem.cross_mode = "DTCS->DTCS"

        if (int(_mem.name[0]) & 0x80) != 0:
            mem.name = _decode_name(_mem.name)

        mem.mode = _mem.isnarrow and "NFM" or "FM"
        mem.skip = pskip and "P" or skip and "S" or ""
        mem.power = POWER_LEVELS[3 - _mem.power]

        return mem

    def set_memory(self, mem):
        _mem = self._memobj.memory[mem.number]
        _flag = self._memobj.flags[(mem.number) / 2]

        nibble = ((mem.number) % 2) and "even" or "odd"

        valid = _flag["%s_valid" % nibble]
        used = _flag["%s_masked" % nibble]

        if not valid:
            _wipe_memory(_mem)

        if mem.empty and valid and not used:
            _flag["%s_valid" % nibble] = False
            return

        _flag["%s_masked" % nibble] = not mem.empty

        if mem.empty:
            return

        _flag["%s_valid" % nibble] = True

        _mem.freq = mem.freq / 1000
        _mem.offset = mem.offset / 1000
        _mem.duplex = DUPLEX.index(mem.duplex)

        # clear all the split tone flags -- we'll set them as needed below
        _mem.ctonesplitflag = 0
        _mem.rx_dtcssplitflag = 0
        _mem.rtonesplitflag = 0
        _mem.dtcssplitflag = 0

        if mem.tmode != "Cross":
            _mem.tmode = TMODES.index(mem.tmode)
            # for the non-cross modes, use ONE tone for both send
            # and receive but figure out where to get it from.
            if mem.tmode == "TSQL" or mem.tmode == "TSQL-R":
                _mem.rtone = chirp_common.TONES.index(mem.ctone)
                _mem.ctone = chirp_common.TONES.index(mem.ctone)
            else:
                _mem.rtone = chirp_common.TONES.index(mem.rtone)
                _mem.ctone = chirp_common.TONES.index(mem.rtone)

            # and one tone for dtcs, but this is always the sending one
            _mem.dtcs = chirp_common.DTCS_CODES.index(mem.dtcs)
            _mem.rx_dtcs = chirp_common.DTCS_CODES.index(mem.dtcs)

        else:
            _mem.rtone = chirp_common.TONES.index(mem.rtone)
            _mem.ctone = chirp_common.TONES.index(mem.ctone)
            _mem.dtcs = chirp_common.DTCS_CODES.index(mem.dtcs)
            _mem.rx_dtcs = chirp_common.DTCS_CODES.index(mem.rx_dtcs)
            if mem.cross_mode == "Tone->Tone":
                # tone->tone cross mode is treated as
                # TSQL, but with separate tones for
                # send and receive
                _mem.tmode = TMODES.index("TSQL")
                _mem.rtonesplitflag = 1
            elif mem.cross_mode == "DTCS->DTCS":
                # DTCS->DTCS cross mode is treated as
                # DTCS, but with separate codes for
                # send and receive
                _mem.tmode = TMODES.index("DTCS")
                _mem.dtcssplitflag = 1
            else:
                _mem.tmode = TMODES.index("Cross") + \
                    CROSS_MODES.index(mem.cross_mode)

        _mem.isnarrow = MODES.index(mem.mode)
        _mem.step = STEPS.index(mem.tuning_step)
        _flag["%s_pskip" % nibble] = mem.skip == "P"
        _flag["%s_skip" % nibble] = mem.skip == "S"
        if mem.power:
            _mem.power = 3 - POWER_LEVELS.index(mem.power)
        else:
            _mem.power = 3

        _mem.name = _encode_name(mem.name)

        # set all unknown areas of the memory map to 0
        _mem.unknown0 = 0
        _mem.unknown1 = 0
        _mem.unknown2 = 0
        _mem.unknown3 = 0
        _mem.unknown4 = 0
        _mem.unknown5 = 0

        LOG.debug("encoded mem\n%s\n" % (util.hexprint(
            _mem.get_raw(asbytes=False)[0:20])))

    def get_settings(self):
        _settings = self._memobj.settings
        _dtmf_strings = self._memobj.dtmf_strings
        _passwd = self._memobj.passwd

        repeater = RadioSettingGroup("repeater", "Repeater Settings")
        ctcss = RadioSettingGroup("ctcss", "CTCSS/DCS/EPCS Settings")
        arts = RadioSettingGroup("arts", "ARTS Settings")
        mbls = RadioSettingGroup("banks", "Memory Settings")
        scan = RadioSettingGroup("scan", "Scan Settings")
        dtmf = RadioSettingGroup("dtmf", "DTMF Settings")
        wires = RadioSettingGroup("wires", "WiRES(tm) Settings")
        switch = RadioSettingGroup("switch", "Switch/Knob Settings")
        disp = RadioSettingGroup("disp", "Display Settings")
        misc = RadioSettingGroup("misc", "Miscellaneous Settings")

        setmode = RadioSettings(repeater, ctcss, arts, mbls, scan,
                                dtmf, wires, switch, disp, misc)

        # numbers and names of settings refer to the way they're
        # presented in the set menu, as well as the list starting on
        # page 74 of the manual

        # 1 APO
        opts = ["Off", "30 Min", "1 Hour", "3 Hour", "5 Hour", "8 Hour"]
        misc.append(
            RadioSetting(
                "apo", "Automatic Power Off",
                RadioSettingValueList(opts, opts[_settings.apo])))

        # 2 AR.BEP
        opts = ["Off", "In Range", "Always"]
        arts.append(
            RadioSetting(
                "arts_beep", "ARTS Beep",
                RadioSettingValueList(opts, opts[_settings.arts_beep])))

        # 3 AR.INT
        opts = ["15 Sec", "25 Sec"]
        arts.append(
            RadioSetting(
                "arts_interval", "ARTS Polling Interval",
                RadioSettingValueList(opts, opts[_settings.arts_interval])))

        # 4 ARS
        opts = ["Off", "On"]
        repeater.append(
            RadioSetting(
                "ars", "Automatic Repeater Shift",
                RadioSettingValueList(opts, opts[_settings.ars])))

        # 5 BCLO
        opts = ["Off", "On"]
        misc.append(RadioSetting(
            "busy_lockout", "Busy Channel Lock-Out",
            RadioSettingValueList(opts, opts[_settings.busy_lockout])))

        # 6 BEEP
        opts = ["Off", "Key+Scan", "Key"]
        switch.append(RadioSetting(
            "beep", "Enable the Beeper",
            RadioSettingValueList(opts, opts[_settings.beep])))

        # 7 BELL
        opts = ["Off", "1", "3", "5", "8", "Continuous"]
        ctcss.append(RadioSetting("bell", "Bell Repetitions",
                                  RadioSettingValueList(opts, opts[
                                                        _settings.bell])))

        # 8 BNK.LNK
        for i in range(0, 8):
            opts = ["Off", "On"]
            mbs = (self._memobj.mbs >> i) & 1
            rs = RadioSetting("mbs%i" % i, "Bank %s Scan" % (i + 1),
                              RadioSettingValueList(opts, opts[mbs]))

            def apply_mbs(s, index):
                if int(s.value):
                    self._memobj.mbs |= (1 << index)
                else:
                    self._memobj.mbs &= ~(1 << index)
            rs.set_apply_callback(apply_mbs, i)
            mbls.append(rs)

        # 9 BNK.NM - A per-bank attribute, nothing to do here.

        # 10 CLK.SFT - A per-channel attribute, nothing to do here.

        # 11 CW.ID
        opts = ["Off", "On"]
        arts.append(RadioSetting("cw_id", "CW ID Enable",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.cw_id])))

        cw_id_text = ""
        for i in _settings.cw_id_string:
            try:
                cw_id_text += CHARSET[i & 0x7F]
            except IndexError:
                if i != 0xff:
                    LOG.debug("unknown char index in cw id: %x " % (i))

        val = RadioSettingValueString(0, 16, cw_id_text, True)
        val.set_charset(CHARSET + "abcdefghijklmnopqrstuvwxyz")
        rs = RadioSetting("cw_id_string", "CW Identifier Text", val)

        def apply_cw_id(s):
            str = s.value.get_value().upper().rstrip()
            mval = ""
            mval = [chr(CHARSET.index(x)) for x in str]
            for x in range(len(mval), 16):
                mval.append(chr(0xff))
            for x in range(0, 16):
                _settings.cw_id_string[x] = ord(mval[x])
        rs.set_apply_callback(apply_cw_id)
        arts.append(rs)

        # 12 CWTRNG
        opts = ["Off", "4WPM", "5WPM", "6WPM", "7WPM", "8WPM", "9WPM",
                "10WPM", "11WPM", "12WPM", "13WPM", "15WPM", "17WPM",
                "20WPM", "24WPM", "30WPM", "40WPM"]
        misc.append(RadioSetting("cw_trng", "CW Training",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.cw_trng])))

        # todo: make the setting of the units here affect the display
        # of the speed.  Not critical, but would be slick.
        opts = ["CPM", "WPM"]
        misc.append(RadioSetting("cw_trng_units", "CW Training Units",
                                 RadioSettingValueList(opts,
                                                       opts[_settings.
                                                            cw_trng_units])))

        # 13 DC VLT - a read-only status, so nothing to do here

        # 14 DCS CD - A per-channel attribute, nothing to do here

        # 15 DCS.RV
        opts = ["Disabled", "Enabled"]
        ctcss.append(RadioSetting(
                     "inverted_dcs",
                     "\"Inverted\" DCS Code Decoding",
                     RadioSettingValueList(opts,
                                           opts[_settings.inverted_dcs])))

        # 16 DIMMER
        opts = ["Off"] + ["Level %d" % (x) for x in range(1, 11)]
        disp.append(RadioSetting("dimmer", "Dimmer",
                                 RadioSettingValueList(opts,
                                                       opts[_settings
                                                            .dimmer])))

        # 17 DT.A/M
        opts = ["Manual", "Auto"]
        dtmf.append(RadioSetting("dtmf_mode", "DTMF Autodialer",
                                 RadioSettingValueList(opts,
                                                       opts[_settings
                                                            .dtmf_mode])))

        # 18 DT.DLY
        opts = ["50 ms", "250 ms", "450 ms", "750 ms", "1000 ms"]
        dtmf.append(RadioSetting("dtmf_delay", "DTMF Autodialer Delay Time",
                                 RadioSettingValueList(opts,
                                                       opts[_settings
                                                            .dtmf_delay])))

        # 19 DT.SET
        for memslot in range(0, 10):
            dtmf_memory = ""
            for i in _dtmf_strings[memslot].dtmf_string:
                if i != 0xFF:
                    try:
                        dtmf_memory += CHARSET[i]
                    except IndexError:
                        LOG.debug("unknown char index in dtmf: %x " % (i))

            val = RadioSettingValueString(0, 16, dtmf_memory, True)
            val.set_charset(CHARSET + "abcdef")
            rs = RadioSetting("dtmf_string_%d" % memslot,
                              "DTMF Memory %d" % memslot, val)

            def apply_dtmf(s, i):
                LOG.debug("applying dtmf for %x\n" % i)
                str = s.value.get_value().upper().rstrip()
                LOG.debug("str is %s\n" % str)
                mval = ""
                mval = [chr(CHARSET.index(x)) for x in str]
                for x in range(len(mval), 16):
                    mval.append(chr(0xff))
                for x in range(0, 16):
                    _dtmf_strings[i].dtmf_string[x] = ord(mval[x])
            rs.set_apply_callback(apply_dtmf, memslot)
            dtmf.append(rs)

        # 20 DT.SPD
        opts = ["50 ms", "100 ms"]
        dtmf.append(RadioSetting("dtmf_speed",
                                 "DTMF Autodialer Sending Speed",
                                 RadioSettingValueList(opts,
                                                       opts[_settings.
                                                            dtmf_speed])))

        # 21 EDG.BEP
        opts = ["Off", "On"]
        mbls.append(RadioSetting("edge_beep", "Band Edge Beeper",
                                 RadioSettingValueList(opts,
                                                       opts[_settings.
                                                            edge_beep])))

        # 22 INT.CD
        opts = ["DTMF %X" % (x) for x in range(0, 16)]
        wires.append(RadioSetting("int_cd", "Access Number for WiRES(TM)",
                                  RadioSettingValueList(opts, opts[
                                                        _settings.int_cd])))

        # 23 ING MD
        opts = ["Sister Radio Group", "Friends Radio Group"]
        wires.append(RadioSetting("wires_mode",
                                  "Internet Link Connection Mode",
                                  RadioSettingValueList(opts,
                                                        opts[_settings.
                                                             wires_mode])))

        # 24 INT.A/M
        opts = ["Manual", "Auto"]
        wires.append(RadioSetting("wires_auto", "Internet Link Autodialer",
                                  RadioSettingValueList(opts,
                                                        opts[_settings
                                                             .wires_auto])))
        # 25 INT.SET
        opts = ["F%d" % (x) for x in range(0, 10)]

        wires.append(RadioSetting("int_set", "Memory Register for "
                                  "non-WiRES Internet",
                                  RadioSettingValueList(opts,
                                                        opts[_settings
                                                             .int_set])))

        # 26 LOCK
        opts = ["Key", "Dial", "Key + Dial", "PTT",
                "Key + PTT", "Dial + PTT", "All"]
        switch.append(RadioSetting("lock", "Control Locking",
                                   RadioSettingValueList(opts,
                                                         opts[_settings
                                                              .lock])))

        # 27 MCGAIN
        opts = ["Level %d" % (x) for x in range(1, 10)]
        misc.append(RadioSetting("mic_gain", "Microphone Gain",
                                 RadioSettingValueList(opts,
                                                       opts[_settings
                                                            .mic_gain])))

        # 28 MEM.SCN
        opts = ["Tag 1", "Tag 2", "All Channels"]
        rs = RadioSetting("scan_mode", "Memory Scan Mode",
                          RadioSettingValueList(opts,
                                                opts[_settings
                                                     .scan_mode - 1]))
        # this setting is unusual in that it starts at 1 instead of 0.
        # that is, index 1 corresponds to "Tag 1", and index 0 is invalid.
        # so we create a custom callback to handle this.

        def apply_scan_mode(s):
            myopts = ["Tag 1", "Tag 2", "All Channels"]
            _settings.scan_mode = myopts.index(s.value.get_value()) + 1
        rs.set_apply_callback(apply_scan_mode)
        mbls.append(rs)

        # 29 MW MD
        opts = ["Lower", "Next"]
        mbls.append(RadioSetting("mw_mode", "Memory Write Mode",
                                 RadioSettingValueList(opts,
                                                       opts[_settings
                                                            .mw_mode])))

        # 30 NM SET - This is per channel, so nothing to do here

        # 31 OPN.MSG
        opts = ["Off", "DC Supply Voltage", "Text Message"]
        disp.append(RadioSetting("open_msg", "Opening Message Type",
                                 RadioSettingValueList(opts,
                                                       opts[_settings.
                                                            open_msg])))

        openmsg = ""
        for i in _settings.openMsg_Text:
            try:
                openmsg += CHARSET[i & 0x7F]
            except IndexError:
                if i != 0xff:
                    LOG.debug("unknown char index in openmsg: %x " % (i))

        val = RadioSettingValueString(0, 6, openmsg, True)
        val.set_charset(CHARSET + "abcdefghijklmnopqrstuvwxyz")
        rs = RadioSetting("openMsg_Text", "Opening Message Text", val)

        def apply_openmsg(s):
            str = s.value.get_value().upper().rstrip()
            mval = ""
            mval = [chr(CHARSET.index(x)) for x in str]
            for x in range(len(mval), 6):
                mval.append(chr(0xff))
            for x in range(0, 6):
                _settings.openMsg_Text[x] = ord(mval[x])
        rs.set_apply_callback(apply_openmsg)
        disp.append(rs)

        # 32 PAGER - a per-channel attribute

        # 33 PAG.ABK
        opts = ["Off", "On"]
        ctcss.append(RadioSetting("pag_abk", "Paging Answer Back",
                                  RadioSettingValueList(opts,
                                                        opts[_settings
                                                             .pag_abk])))

        # 34 PAG.CDR
        opts = ["%2.2d" % (x) for x in range(1, 50)]
        ctcss.append(RadioSetting("pag_cdr_1", "Receive Page Code 1",
                                  RadioSettingValueList(opts,
                                                        opts[_settings
                                                             .pag_cdr_1])))

        ctcss.append(RadioSetting("pag_cdr_2", "Receive Page Code 2",
                                  RadioSettingValueList(opts,
                                                        opts[_settings
                                                             .pag_cdr_2])))

        # 35 PAG.CDT
        opts = ["%2.2d" % (x) for x in range(1, 50)]
        ctcss.append(RadioSetting("pag_cdt_1", "Transmit Page Code 1",
                                  RadioSettingValueList(opts,
                                                        opts[_settings
                                                             .pag_cdt_1])))

        ctcss.append(RadioSetting("pag_cdt_2", "Transmit Page Code 2",
                                  RadioSettingValueList(opts,
                                                        opts[_settings
                                                             .pag_cdt_2])))

        # Common Button Options
        button_opts = ["Squelch Off", "Weather", "Smart Search",
                       "Tone Scan", "Scan", "T Call", "ARTS"]

        # 36 PRG P1
        opts = button_opts + ["DC Volts"]
        switch.append(RadioSetting(
            "prog_p1", "P1 Button",
            RadioSettingValueList(opts, opts[_settings.prog_p1])))

        # 37 PRG P2
        opts = button_opts + ["Dimmer"]
        switch.append(RadioSetting(
            "prog_p2", "P2 Button",
            RadioSettingValueList(opts, opts[_settings.prog_p2])))

        # 38 PRG P3
        opts = button_opts + ["Mic Gain"]
        switch.append(RadioSetting(
            "prog_p3", "P3 Button",
            RadioSettingValueList(opts, opts[_settings.prog_p3])))

        # 39 PRG P4
        opts = button_opts + ["Skip"]
        switch.append(RadioSetting(
            "prog_p4", "P4 Button",
            RadioSettingValueList(opts, opts[_settings.prog_p4])))

        # 40 PSWD
        password = ""
        for i in _passwd:
            if i != 0xFF:
                try:
                    password += CHARSET[i]
                except IndexError:
                    LOG.debug("unknown char index in password: %x " % (i))

        val = RadioSettingValueString(0, 4, password, True)
        val.set_charset(CHARSET[0:15] + "abcdef ")
        rs = RadioSetting("passwd", "Password", val)

        def apply_password(s):
            str = s.value.get_value().upper().rstrip()
            mval = ""
            mval = [chr(CHARSET.index(x)) for x in str]
            for x in range(len(mval), 4):
                mval.append(chr(0xff))
            for x in range(0, 4):
                _passwd[x] = ord(mval[x])
        rs.set_apply_callback(apply_password)
        misc.append(rs)

        # 41 RESUME
        opts = ["3 Sec", "5 Sec", "10 Sec", "Busy", "Hold"]
        scan.append(RadioSetting("resume", "Scan Resume Mode",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.resume])))

        # 42 RF.SQL
        opts = ["Off"] + ["S-%d" % (x) for x in range(1, 10)]
        misc.append(RadioSetting("rf_sql", "RF Squelch Threshold",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.rf_sql])))

        # 43 RPT - per channel attribute, nothing to do here

        # 44 RVRT
        opts = ["Off", "On"]
        misc.append(RadioSetting("revert", "Priority Revert",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.revert])))

        # 45 S.SRCH
        opts = ["Single", "Continuous"]
        misc.append(RadioSetting("s_search", "Smart Search Sweep Mode",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.s_search])))

        # 46 SHIFT - per channel setting, nothing to do here

        # 47 SKIP = per channel setting, nothing to do here

        # 48 SPLIT - per channel attribute, nothing to do here

        # 49 SQL.TYP - per channel attribute, nothing to do here

        # 50 STEP - per channel attribute, nothing to do here

        # 51 TEMP - read-only status, nothing to do here

        # 52 TN FRQ - per channel attribute, nothing to do here

        # 53 TOT
        opts = ["Off", "1 Min", "3 Min", "5 Min", "10 Min"]
        misc.append(RadioSetting("tot", "Timeout Timer",
                                 RadioSettingValueList(opts,
                                                       opts[_settings.tot])))

        # 54 TS MUT
        opts = ["Off", "On"]
        ctcss.append(RadioSetting("ts_mut", "Tone Search Mute",
                                  RadioSettingValueList(opts,
                                                        opts[_settings
                                                             .ts_mut])))

        # 55 TS SPEED
        opts = ["Fast", "Slow"]
        ctcss.append(RadioSetting("ts_speed", "Tone Search Scanner Speed",
                                  RadioSettingValueList(opts,
                                                        opts[_settings
                                                             .ts_speed])))

        # 56 VFO.SCN
        opts = ["+/- 1 MHz", "+/- 2 MHz", "+/-5 MHz", "All"]
        scan.append(RadioSetting("vfo_scan", "VFO Scanner Width",
                                 RadioSettingValueList(opts,
                                                       opts[_settings
                                                            .vfo_scan])))

        # 57 WX.ALT
        opts = ["Off", "On"]
        misc.append(RadioSetting("wx_alert", "Weather Alert Scan",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.wx_alert])))

        # 58 WX.VOL
        opts = ["Normal", "Maximum"]
        misc.append(RadioSetting("wx_vol_max", "Weather Alert Volume",
                                 RadioSettingValueList(opts, opts[
                                                       _settings.wx_vol_max])))

        # 59 W/N DV - this is a per-channel attribute, nothing to do here

        return setmode

    def set_settings(self, uisettings):
        _settings = self._memobj.settings
        for element in uisettings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            if not element.changed():
                continue

            try:
                name = element.get_name()
                value = element.value

                if element.has_apply_callback():
                    LOG.debug("Using apply callback")
                    element.run_apply_callback()
                else:
                    obj = getattr(_settings, name)
                    setattr(_settings, name, value)

                LOG.debug("Setting %s: %s" % (name, value))
            except Exception:
                LOG.debug(element.get_name())
                raise

    def get_bank_model(self):
        return FT2900BankModel(self)

    @classmethod
    def match_model(cls, filedata, filename):
        return len(filedata) == cls._memsize

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.pre_download = _(
            "1. Turn Radio off.\n"
            "2. Connect data cable.\n"
            "3. While holding \"A/N LOW\" button, turn radio on.\n"
            "4. <b>After clicking OK</b>, press \"SET MHz\" to send image.\n")
        rp.pre_upload = _(
            "1. Turn Radio off.\n"
            "2. Connect data cable.\n"
            "3. While holding \"A/N LOW\" button, turn radio on.\n"
            "4. Press \"MW D/MR\" to receive image.\n"
            "5. Make sure display says \"-WAIT-\" (see note below if not)\n"
            "6. Click OK to dismiss this dialog and start transfer.\n"
            "Note: if you don't see \"-WAIT-\" at step 5, try cycling\n"
            "      power and pressing and holding red \"*L\" button to"
            " unlock\n"
            "      radio, then start back at step 1.\n")
        return rp


# the FT2900E is the European version of the radio, almost identical
# to the R (USA) version, except for the model number and ID Block.  We
# create and register a class for it, with only the needed overrides
# NOTE: Disabled until detection is fixed
# @directory.register
class FT2900ERadio(FT2900Radio):

    """Yaesu FT-2900E"""
    MODEL = "FT-2900E/1900E"
    VARIANT = "E"
    IDBLOCK = b"\x56\x43\x32\x33\x00\x02\x41\x02\x01\x01"


# This class is for the TX Modified FT-1900/FT-2900 (MARS/CAP Mod).
# Enabling out of band TX changes the header received by CHIRP
@directory.register
class FT2900ModRadio(FT2900Radio):

    """Yaesu FT-2900Mod"""
    MODEL = "FT-2900R/1900R(TXMod)"
    VARIANT = "Opened Xmit"
    IDBLOCK = b"\x56\x43\x32\x33\x00\x02\xc7\x01\x01\x01"

    @classmethod
    def match_model(cls, filedata, filename):
        return False
